# ***************************************************************************
# *   Copyright (c) 2017 Markus Hovorka <m.hovorka@live.de>                 *
# *   Copyright (c) 2017 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Execute Solver and obtain Reports and Results.

Integral part of the Solver Framework which contains components responsible for
executing the solver in the background. Also provides an asynchronous
communication system with the solver running in the background. The purpose of
this module is to be as generic as possible. It can execute every solver
supported by the fem workbench. The threading and communication support is
mainly implemented by the :mod:`femsolver.task` and :mod:`femsolver.signal`
modules.
"""

__title__ = "FreeCAD FEM solver run"
__author__ = "Markus Hovorka, Bernd Hahnebach"
__url__ = "https://www.freecad.org"

import os
import os.path
import shutil
import tempfile
from PySide import QtCore

# import threading  # not used ATM

import FreeCAD as App

from . import settings
from . import signal
from . import task
from femsolver.elmer import elmertools
from femsolver.calculix import calculixtools
from femtools import femutils
from femtools import membertools
from femtools.errors import DirectoryDoesNotExistError
from femtools.errors import MustSaveError

if App.GuiUp:
    from PySide import QtGui
    import FreeCADGui


CHECK = 0
PREPARE = 1
SOLVE = 2
RESULTS = 3
DONE = 4


_machines = {}
_dirTypes = {}


def run_fem_solver(solver, working_dir=None, blocking=False):
    """Execute *solver* of the solver framework.

    Uses :meth:`getMachine <femsolver.solverbase.Proxy.getMachine>` to obtain a
    :class:`Machine` instance of the solver. It than executes the Machine with
    using the ``RESULTS`` target (see :class:`Machine` for infos about
    different targets). This method is blocking, it waits for the solver to
    finished before returning. Be aware of :class:`Machine` caching when using
    the function.

    :param solver:
        A document object which must be a framework compliant solver. This means
        that it should be derived from the document object provided by
        :mod:`femsolver.solverbase` and implement all required methods
        correctly. Of particular importance is :meth:`getMachine
        <femsolver.solverbase.Proxy.getMachine>` as it is used by this method
        the get the :class:`Machine` used to execute the solver.

    :param working_dir:
        If specified it overwrites the automatic and user configurable working
        directory management of the Solver framework. Should always be a
        absolute path because the location of the binary is not consistent
        among platforms. If ``None`` the automatic working directory management
        is used.

    :note:
        There is some legacy code to execute the old Calculix solver
        (pre-framework) which behaives differently because it does not
        use a :class:`Machine`.
    """

    tool = None
    if working_dir:
        solver.WorkingDirectory = working_dir

    match solver.Proxy.Type:
        case "Fem::SolverElmer":
            tool = elmertools.ElmerTools(solver)
        case "Fem::SolverCalculiX":
            tool = calculixtools.CalculiXTools(solver)

    if tool is not None:
        # Redirect process error to report view
        print_error = lambda: App.Console.PrintError(
            tool.process.readAllStandardError().data().decode("utf-8")
        )
        tool.process.readyReadStandardError.connect(print_error)
        tool.process.finished.connect(_solver_finish(solver))
        try:
            if App.GuiUp:
                QtGui.QApplication.setOverrideCursor(QtCore.Qt.WaitCursor)
            tool.run(blocking)
        except Exception as e:
            if App.GuiUp:
                QtGui.QApplication.restoreOverrideCursor()
            raise e
        return

    # code for old solver implementations

    if solver.Proxy.Type == "Fem::SolverCcxTools":
        from femtools.ccxtools import CcxTools as ccx

        App.Console.PrintMessage("Run of CalxuliX ccx tools solver started.\n")
        fea = ccx(solver)
        fea.reset_mesh_purge_results_checked()
        if working_dir is None:
            fea.run()  # standard, no working dir is given in solver
        else:
            # not the standard way
            fea.update_objects()
            fea.setup_working_dir(working_dir)
            fea.setup_ccx()
            message = fea.check_prerequisites()
            if not message:
                fea.write_inp_file()
                fea.ccx_run()
                fea.load_results()
            else:
                App.Console.PrintError(f"Houston, we have a problem...!\n{message}\n")
        App.Console.PrintMessage("Run of CalxuliX ccx tools solver finished.\n")
    else:
        # App.Console.PrintMessage("Frame work solver!\n")
        try:
            if working_dir is not None:
                machine = getMachine(solver, working_dir)
            else:
                machine = getMachine(solver)
        except MustSaveError:
            error_message = (
                "Please save the file before executing the solver. "
                "This must be done because the location of the working "
                'directory is set to "Beside *.FCStd File".'
            )
            App.Console.PrintError(error_message + "\n")
            if App.GuiUp:
                QtGui.QMessageBox.critical(
                    FreeCADGui.getMainWindow(), "Can't start Solver", error_message
                )
            return
        except DirectoryDoesNotExistError:
            error_message = "Selected working directory doesn't exist."
            App.Console.PrintError(error_message + "\n")
            if App.GuiUp:
                QtGui.QMessageBox.critical(
                    FreeCADGui.getMainWindow(), "Can't start Solver", error_message
                )
            return
        if not machine.running:
            machine.reset()
            machine.target = RESULTS
            machine.start()
            machine.join()  # wait for the machine to finish.
            if machine.failed is True:
                App.Console.PrintError("Machine failed to run.\n")
                from .report import displayLog

                displayLog(machine.report)
                if App.GuiUp:
                    error_message = (
                        "Failed to run. Please try again after all "
                        "of the following errors are resolved."
                    )
                    from .report import display

                    display(machine.report, "Run Report", error_message)


def getMachine(solver, path=None):
    """Get or create :class:`Machine` using caching mechanism.

    :param solver:
        A document object which must be a framework compliant solver. This means
        that it should be derived from the document object provided by
        :mod:`femsolver.solverbase` and implement all required methods
        correctly. Of particular importance is :meth:`getMachine
        <femsolver.solverbase.Proxy.getMachine>` as it is used by this method
        to create a new :class:`Machine` on cache miss.

    :param path:
        A valid filesystem path which shall be associetad with the machine.
    """
    # print(path)
    _DocObserver.attach()
    m = _machines.get(solver)
    if m is None or not _isPathValid(m, path):
        m = _createMachine(solver, path, testmode=False)
        # print(m.__dir__())  # document these attributes somewhere
        # print(m.directory)
    return m


def _solver_finish(obj):
    def receiver(code, status):
        if status != QtCore.QProcess.ExitStatus.NormalExit or code != 0:
            App.Console.PrintError("Solver finished with errors. Result not updated\n")
        if App.GuiUp:
            QtGui.QApplication.restoreOverrideCursor()
        obj.Document.recompute()

    return receiver


def _isPathValid(m, path):
    t = _dirTypes.get(m.directory)  # setting default None
    setting = settings.get_dir_setting()
    if path is not None:
        return t is None and m.directory == path
    if setting == settings.DirSetting.BESIDE:
        if t == settings.DirSetting.BESIDE:
            base = os.path.split(m.directory.rstrip("/"))[0]
            return base == _getBesideBase(m.solver)
        return False
    if setting == settings.DirSetting.TEMPORARY:
        return t == settings.DirSetting.TEMPORARY
    if setting == settings.DirSetting.CUSTOM:
        if t == settings.DirSetting.CUSTOM:
            firstBase = os.path.split(m.directory.rstrip("/"))[0]
            customBase = os.path.split(firstBase)[0]
            return customBase == _getCustomBase(m.solver)
        return False


def _createMachine(solver, path, testmode):
    global _dirTypes
    setting = settings.get_dir_setting()
    if path is not None:
        _dirTypes[path] = None
    elif setting == settings.DirSetting.BESIDE:
        path = _getBesideDir(solver)
        _dirTypes[path] = settings.DirSetting.BESIDE
    elif setting == settings.DirSetting.TEMPORARY:
        path = _getTempDir(solver)
        _dirTypes[path] = settings.DirSetting.TEMPORARY
    elif setting == settings.DirSetting.CUSTOM:
        path = _getCustomDir(solver)
        _dirTypes[path] = settings.DirSetting.CUSTOM
    m = solver.Proxy.createMachine(solver, path, testmode)
    oldMachine = _machines.get(solver)
    if oldMachine is not None and _dirTypes.get(oldMachine.directory) is not None:
        del _dirTypes[oldMachine.directory]
    _machines[solver] = m
    return m


def _getTempDir(solver):
    return tempfile.mkdtemp(prefix="fem")


def _getBesideDir(solver):
    base = _getBesideBase(solver)
    specificPath = os.path.join(base, solver.Label)
    specificPath = _getUniquePath(specificPath)
    if not os.path.isdir(specificPath):
        os.makedirs(specificPath)
    return specificPath


def _getBesideBase(solver):
    path = os.path.splitext(solver.Document.FileName)[0]
    # doc=App.newDocument()
    # doc.FileName
    # the above returns an empty string in FreeCAD 0.19
    # https://forum.freecad.org/viewtopic.php?f=10&t=48842
    if path == "":
        error_message = (
            "Please save the file before executing the solver. "
            "This must be done because the location of the working "
            'directory is set to "Beside *.FCStd File".'
        )
        App.Console.PrintError(error_message + "\n")
        if App.GuiUp:
            QtGui.QMessageBox.critical(
                FreeCADGui.getMainWindow(), "Can't start Solver", error_message
            )
        raise MustSaveError()
        # TODO may be do not abort but use a temporary directory
    return path


def _getCustomDir(solver):
    base = _getCustomBase(solver)
    specificPath = os.path.join(base, solver.Document.Name, solver.Label)
    specificPath = _getUniquePath(specificPath)
    if not os.path.isdir(specificPath):
        os.makedirs(specificPath)
    return specificPath


def _getCustomBase(solver):
    path = settings.get_custom_dir()
    if not os.path.isdir(path):
        error_message = "Selected working directory doesn't exist."
        App.Console.PrintError(error_message + "\n")
        if App.GuiUp:
            QtGui.QMessageBox.critical(
                FreeCADGui.getMainWindow(), "Can't start Solver", error_message
            )
        raise DirectoryDoesNotExistError("Invalid path")
    return path


def _getUniquePath(path):
    postfix = 1
    if path in _dirTypes:
        path += "_%03d" % postfix
    while path in _dirTypes:
        postfix += 1
        path = path[:-4] + "_%03d" % postfix
    return path


class BaseTask(task.Thread):

    def __init__(self):
        super().__init__()
        self.solver = None
        self.directory = None
        self.testmode = None

    @property
    def analysis(self):
        return self.solver.getParentGroup()


class Machine(BaseTask):

    def __init__(self, solver, directory, check, prepare, solve, results, testmode):
        super().__init__()
        self.solver = solver
        self.directory = directory
        self.signalState = set()
        self.check = check
        self.prepare = prepare
        self.solve = solve
        self.results = results
        self.target = RESULTS
        self._state = CHECK
        self._pendingState = None
        self._isReset = False
        self.testmode = testmode

    @property
    def state(self):
        return self._state

    def run(self):
        self._confTasks()
        self._isReset = False
        self._pendingState = self.state
        while not self.aborted and not self.failed and self._pendingState <= self.target:
            task = self._getTask(self._pendingState)
            self._runTask(task)
            self.report.extend(task.report)
            if task.failed:
                self.fail()
            elif task.aborted:
                self.abort()
            else:
                self._pendingState += 1
        self._applyPending()

    def reset(self, newState=CHECK):
        state = self.state if self._pendingState is None else self._pendingState
        if newState < state:
            self._isReset = True
            self._state = newState
            signal.notify(self.signalState)

    def _confTasks(self):
        tasks = [self.check, self.prepare, self.solve, self.results]
        for t in tasks:
            t.solver = self.solver
            t.directory = self.directory
            t.testmode = self.testmode

    def _applyPending(self):
        if not self._isReset:
            self._state = self._pendingState
            signal.notify(self.signalState)
        self._isReset = False
        self._pendingState = None

    def _runTask(self, task):

        def statusProxy(line):
            self.pushStatus(line)

        def killer():
            task.abort()

        self.signalAbort.add(killer)
        task.signalStatus.add(statusProxy)
        task.start()
        task.join()
        self.signalAbort.remove(killer)
        task.signalStatus.remove(statusProxy)

    def _getTask(self, state):
        if state == CHECK:
            return self.check
        elif state == PREPARE:
            return self.prepare
        elif state == SOLVE:
            return self.solve
        elif state == RESULTS:
            return self.results
        return None


class Check(BaseTask):

    def get_several_member(self, t):
        return membertools.get_several_member(self.analysis, t)

    def check_mesh_exists(self):
        meshes = self.get_several_member("Fem::FemMeshObject")
        if len(meshes) == 0:
            self.report.error("Missing a mesh object.")
            self.fail()
            return False
        elif len(meshes) > 1:
            self.report.error("Too many meshes. More than one mesh is not supported.")
            self.fail()
            return False
        return True

    def check_material_exists(self):
        objs = self.get_several_member("App::MaterialObjectPython")
        if len(objs) == 0:
            self.report.error("Missing a material object. At least one material is required.")
            self.fail()
            return False
        return True

    def check_material_single(self):
        objs = self.get_several_member("App::MaterialObjectPython")
        if len(objs) > 1:
            self.report.error("Only one Material is supported for this solver.")
            self.fail()
            return False
        return True

    def check_geos_beamsection_no(self):
        objs = self.get_several_member("Fem::ElementGeometry1D")
        if len(objs) > 0:
            self.report.error("Beamsections are not supported for this solver.")
            self.fail()
            return False
        return True

    def check_geos_beamsection_single(self):
        objs = self.get_several_member("Fem::ElementGeometry1D")
        if len(objs) > 1:
            self.report.error("Only one beamsection is supported for this solver.")
            self.fail()
            return False
        return True

    def check_geos_shellthickness_no(self):
        objs = self.get_several_member("Fem::ElementGeometry2D")
        if len(objs) > 0:
            self.report.error("Shellsections are not supported for this solver.")
            self.fail()
            return False
        return True

    def check_geos_shellthickness_single(self):
        objs = self.get_several_member("Fem::ElementGeometry2D")
        if len(objs) > 1:
            self.report.error("Only one shellthickness is supported for this solver.")
            self.fail()
            return False
        return True

    def check_geos_beamsection_and_shellthickness(self):
        beamsec_obj = self.get_several_member("Fem::ElementGeometry1D")
        shellth_obj = self.get_several_member("Fem::ElementGeometry2D")
        if len(beamsec_obj) > 0 and len(shellth_obj) > 0:
            self.report.error(
                "Either beamsection or shellthickness objects are "
                "supported for this solver, but not both in one analysis."
            )
            self.fail()
            return False
        return True

    def checkSupported(self, allSupported):
        for m in self.analysis.Group:
            if femutils.is_of_type(m, "Fem::Constraint"):
                supported = False
                for sc in allSupported:
                    if femutils.is_of_type(m, *sc):
                        supported = True
                if not supported:
                    self.report.warning(f"Ignored unsupported constraint: {m.Label}")
        return True


class Solve(BaseTask):

    def _observeSolver(self, process):
        output = ""
        line = femutils.pydecode(process.stdout.readline())
        self.pushStatus(line)
        output += line
        line = femutils.pydecode(process.stdout.readline())
        while line:
            line = "\n%s" % line.rstrip()
            self.pushStatus(line)
            output += line
            line = femutils.pydecode(process.stdout.readline())
        return output


class Prepare(BaseTask):
    pass


class Results(BaseTask):
    pass


class _DocObserver:

    _instance = None
    _WHITELIST = [
        "Fem::Constraint",
        "App::MaterialObject",
        "Fem::FemMeshObject",
    ]
    _BLACKLIST_PROPS = ["Label", "ElmerOutput", "ElmerResult"]

    def __init__(self):
        self._saved = {}
        for doc in iter(App.listDocuments().values()):
            for obj in doc.Objects:
                if obj.isDerivedFrom("Fem::FemAnalysis"):
                    self._saved[obj] = obj.Group

    @classmethod
    def attach(cls):
        if cls._instance is None:
            cls._instance = cls()
            App.addDocumentObserver(cls._instance)

    def slotDeletedObject(self, obj):
        self._checkModel(obj)
        if obj in _machines:
            self._deleteMachine(obj)

    def slotChangedObject(self, obj, prop):
        if prop not in self._BLACKLIST_PROPS:
            self._checkAnalysis(obj)
            self._checkEquation(obj)
            self._checkSolver(obj)
            self._checkModel(obj)

    def slotDeletedDocument(self, doc):
        for obj in doc.Objects:
            if obj in _machines:
                self._deleteMachine(obj)

    def _deleteMachine(self, obj):
        m = _machines[obj]
        t = _dirTypes[m.directory]
        m.abort()
        if t == settings.DirSetting.TEMPORARY:
            shutil.rmtree(m.directory)
        del _machines[obj]
        del _dirTypes[m.directory]

    def _checkEquation(self, obj):
        for o in obj.Document.Objects:
            if (
                femutils.is_derived_from(o, "Fem::FemSolverObject")
                and hasattr(o, "Group")
                and obj in o.Group
            ):
                if o in _machines:
                    _machines[o].reset()

    def _checkSolver(self, obj):
        analysis = obj.getParentGroup()
        for m in iter(_machines.values()):
            if analysis == m.analysis and obj == m.solver:
                m.reset()

    def _checkAnalysis(self, obj):
        if femutils.is_derived_from(obj, "Fem::FemAnalysis"):
            deltaObjs = self._getAdded(obj)
            if deltaObjs:
                reset = False
                for o in deltaObjs:
                    if self._partOfModel(o):
                        reset = True
                if reset:
                    self._resetAll(obj)

    def _checkModel(self, obj):
        if self._partOfModel(obj):
            analysis = obj.getParentGroup()
            if analysis is not None:
                self._resetAll(analysis)

    def _getAdded(self, analysis):
        if analysis not in self._saved:
            self._saved[analysis] = []
        delta = set(analysis.Group) - set(self._saved[analysis])
        self._saved[analysis] = analysis.Group
        return delta

    def _resetAll(self, analysis):
        for m in iter(_machines.values()):
            if analysis == m.analysis:
                m.reset()

    def _partOfModel(self, obj):
        for t in self._WHITELIST:
            if femutils.is_derived_from(obj, t):
                return True
        return False



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the object code for the Text object."""
## @package text
# \ingroup draftobjects
# \brief Provides the object code for the Text object.

## \addtogroup draftobjects
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
from draftobjects.draft_annotation import DraftAnnotation
from draftutils import gui_utils
from draftutils.messages import _log


class Text(DraftAnnotation):
    """The Draft Text object."""

    def __init__(self, obj):
        obj.Proxy = self
        self.Type = "Text"
        self.set_properties(obj)

    def set_properties(self, obj):
        """Add properties to the object and set them."""
        properties = obj.PropertiesList

        if "Placement" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property", "The placement of the base point " "of the first line"
            )
            obj.addProperty("App::PropertyPlacement", "Placement", "Base", _tip, locked=True)
            obj.Placement = App.Placement()

        if "Text" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "The text displayed by this object.\n"
                "It is a list of strings; each element "
                "in the list will be displayed "
                "in its own line.",
            )
            obj.addProperty("App::PropertyStringList", "Text", "Base", _tip, locked=True)
            obj.Text = []

    def onDocumentRestored(self, obj):
        """Execute code when the document is restored."""
        super().onDocumentRestored(obj)
        gui_utils.restore_view_object(obj, vp_module="view_text", vp_class="ViewProviderText")

        vobj = getattr(obj, "ViewObject", None)
        if vobj is None:
            return

        # See loads:
        if self.stored_type is not None:
            self.update_properties_0v21(obj, vobj)

        if hasattr(vobj, "LineWidth") or hasattr(vobj, "LineColor"):
            self.update_properties_1v1(obj, vobj)

    def update_properties_0v21(self, obj, vobj):
        """Update view properties."""
        # The DisplayMode is updated automatically but the new values are
        # switched: "2D text" becomes "World" and "3D text" becomes "Screen".
        # It should be the other way around:
        vobj.DisplayMode = "World" if vobj.DisplayMode == "Screen" else "Screen"
        _log("v0.21, " + obj.Name + ", renamed 'DisplayMode' options to 'World/Screen'")

    def update_properties_1v1(self, obj, vobj):
        if hasattr(vobj, "LineWidth"):
            vobj.setPropertyStatus("LineWidth", "-LockDynamic")
            vobj.removeProperty("LineWidth")
        if hasattr(vobj, "LineColor"):
            vobj.setPropertyStatus("LineColor", "-LockDynamic")
            vobj.removeProperty("LineColor")
        _log("v1.1, " + obj.Name + ", removed view properties")

    def loads(self, state):
        # Before update_properties_0v21 the self.Type value was stored.
        # We use this to identify older objects that need to be updated.
        self.stored_type = state
        self.Type = "Text"


# Alias for compatibility with v0.18 and earlier
DraftText = Text

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2024 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""BIM Schedule command"""

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate

PARAMS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/BIM")


class Arch_Stairs:
    "the Arch Stairs command definition"

    def GetResources(self):

        return {
            "Pixmap": "Arch_Stairs",
            "MenuText": QT_TRANSLATE_NOOP("Arch_Stairs", "Stairs"),
            "Accel": "S, R",
            "ToolTip": QT_TRANSLATE_NOOP("Arch_Stairs", "Creates a flight of stairs"),
        }

    def IsActive(self):

        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):

        import Draft
        from draftutils import params

        FreeCAD.ActiveDocument.openTransaction(translate("Arch", "Create Stairs"))
        FreeCADGui.addModule("Arch")
        sel = FreeCADGui.Selection.getSelection()
        if sel:
            n = []
            nStr = ""
            for obj in sel:
                if nStr != "":
                    nStr += ","
                nStr += "FreeCAD.ActiveDocument." + obj.Name
            #'obj' in GUI not the same as obj in script,
            # make it 'stairs' to distinguish one from another
            # Create Stairs object with steps numbers in user preference
            FreeCADGui.doCommand(
                "stairs = Arch.makeStairs(baseobj=["
                + nStr
                + "],steps="
                + str(params.get_param_arch("StairsSteps"))
                + ")"
            )
            FreeCADGui.Selection.clearSelection()
            FreeCADGui.doCommand("FreeCADGui.Selection.addSelection(stairs)")

            # ArchSketch Support
            if len(sel) == 1 and Draft.getType(obj) == "ArchSketch":
                # Get ArchSketch.FloorHeight as default and assign to Stairs
                try:
                    height = str(obj.FloorHeight.Value)  # vs obj.FloorHeight
                    # Can only use Value to assign to PropertyLength
                    FreeCADGui.doCommand("stairs.Height = " + height)
                except:
                    pass
                # If base is ArchSketch, ArchSketchObject is already loaded, no
                # need to load again : FreeCADGui.addModule("ArchSketchObject")
                try:
                    FreeCADGui.runCommand("EditStairs")
                except:
                    pass

        else:
            FreeCADGui.doCommand(
                "stairs = Arch.makeStairs(steps=" + str(params.get_param_arch("StairsSteps")) + ")"
            )
        FreeCADGui.addModule("Draft")

        # FreeCADGui.doCommand("Draft.autogroup(obj)")
        FreeCADGui.doCommand("Draft.autogroup(stairs)")

        FreeCAD.ActiveDocument.commitTransaction()
        FreeCAD.ActiveDocument.recompute()
        print(" ActiveDocument.recompute, done ")


FreeCADGui.addCommand("Arch_Stairs", Arch_Stairs())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2018 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""This module contains FreeCAD commands for the BIM workbench"""

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate


class BIM_IfcElements:

    def GetResources(self):
        return {
            "Pixmap": "BIM_IfcElements",
            "MenuText": QT_TRANSLATE_NOOP("BIM_IfcElements", "Manage IFC Elements"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "BIM_IfcElements",
                "Manages how the different elements of the BIM project will be exported to IFC",
            ),
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):

        # only raise the dialog if it is already open
        if getattr(self, "form", None):
            self.form.raise_()
            return

        import Draft
        from PySide import QtGui

        # build objects list
        self.objectslist = {}
        try:
            import ArchIFC

            self.ifctypes = ArchIFC.IfcTypes
        except (ImportError, AttributeError):
            import ArchComponent

            self.ifctypes = ArchComponent.IfcRoles
        for obj in FreeCAD.ActiveDocument.Objects:
            mat = ""
            role = self.getRole(obj)
            if role:
                try:
                    mat = obj.Material.Name
                except AttributeError:
                    mat = ""
                self.objectslist[obj.Name] = [role, mat]

        # load the form and set the tree model up
        self.form = FreeCADGui.PySideUic.loadUi(":/ui/dialogIfcElements.ui")
        self.form.setWindowIcon(QtGui.QIcon(":/icons/BIM_IfcElements.svg"))
        self.model = QtGui.QStandardItemModel()
        self.form.tree.setModel(self.model)
        self.form.tree.setUniformRowHeights(True)
        self.form.tree.setItemDelegate(IfcElementsDelegate(dialog=self))
        self.form.globalMode.addItems([" "] + self.ifctypes)
        self.form.groupMode.setItemIcon(2, QtGui.QIcon(":/icons/Arch_Material.svg"))
        self.form.groupMode.setItemIcon(3, QtGui.QIcon(":/icons/Document.svg"))
        self.form.globalMaterial.addItem(" ")
        self.form.globalMaterial.addItem(translate("BIM", "Create new material"))
        self.form.globalMaterial.addItem(translate("BIM", "Create new multi-material"))
        self.materials = []
        for o in FreeCAD.ActiveDocument.Objects:
            if o.isDerivedFrom("App::MaterialObject") or (Draft.getType(o) == "MultiMaterial"):
                self.materials.append(o.Name)
                self.form.globalMaterial.addItem(o.Label, QtGui.QIcon(":/icons/Arch_Material.svg"))
        self.form.groupMode.currentIndexChanged.connect(self.update)
        self.form.tree.clicked.connect(self.onClickTree)
        if hasattr(self.form.onlyVisible, "checkStateChanged"):  # Qt version >= 6.7.0
            self.form.onlyVisible.checkStateChanged.connect(self.update)
        else:  # Qt version < 6.7.0
            self.form.onlyVisible.stateChanged.connect(self.update)
        self.form.buttonBox.accepted.connect(self.accept)
        self.form.rejected.connect(self.reject)  # also triggered by self.form.buttonBox.rejected
        self.form.globalMode.currentIndexChanged.connect(self.onObjectTypeChanged)
        self.form.globalMaterial.currentIndexChanged.connect(self.onMaterialChanged)

        # center the dialog over FreeCAD window
        mw = FreeCADGui.getMainWindow()
        self.form.move(
            mw.frameGeometry().topLeft() + mw.rect().center() - self.form.rect().center()
        )

        self.update()
        self.form.show()

    def update(self, index=None):
        "updates the tree widgets in all tabs"

        # store current state of tree into self.objectslist before redrawing
        for row in range(self.model.rowCount()):
            name = self.model.item(row, 0).toolTip()
            mat = self.model.item(row, 2).toolTip()
            if name:
                self.objectslist[name] = [self.model.item(row, 1).text(), mat]
            if self.model.item(row, 0).hasChildren():
                for childrow in range(self.model.item(row, 0).rowCount()):
                    name = self.model.item(row, 0).child(childrow, 0).toolTip()
                    mat = self.model.item(row, 0).child(childrow, 2).toolTip()
                    if name:
                        self.objectslist[name] = [
                            self.model.item(row, 0).child(childrow, 1).text(),
                            mat,
                        ]
        self.model.clear()
        self.model.setHorizontalHeaderLabels(
            [
                translate("BIM", "Label"),
                translate("BIM", "IFC type"),
                translate("BIM", "Material"),
            ]
        )
        # self.form.tree.header().setResizeMode(QtGui.QHeaderView.Stretch)
        # self.form.tree.resizeColumnsToContents()

        if self.form.groupMode.currentIndex() == 1:
            # group by type
            self.updateByType()
        elif self.form.groupMode.currentIndex() == 2:
            # group by material
            self.updateByMaterial()
        elif self.form.groupMode.currentIndex() == 3:
            # group by model structure
            self.updateByTree()
        else:
            # group alphabetically
            self.updateDefault()
        self.model.sort(0)

    def updateByType(self):
        groups = {}
        for name, rolemat in self.objectslist.items():
            role = rolemat[0]
            mat = rolemat[1]
            obj = FreeCAD.ActiveDocument.getObject(name)
            if obj:
                if (not self.form.onlyVisible.isChecked()) or obj.ViewObject.isVisible():
                    groups.setdefault(role, []).append([name, mat])
        for group in groups.keys():
            s1 = group + " (" + str(len(groups[group])) + ")"
            top = QtGui.QStandardItem(s1)
            self.model.appendRow([top, QtGui.QStandardItem(), QtGui.QStandardItem()])
            for name, mat in groups[group]:
                obj = FreeCAD.ActiveDocument.getObject(name)
                if obj:
                    it1 = QtGui.QStandardItem(obj.Label)
                    it1.setIcon(getIcon(obj))
                    it1.setToolTip(obj.Name)
                    it2 = QtGui.QStandardItem(group)
                    if group != self.getRole(obj):
                        it2.setIcon(QtGui.QIcon(":/icons/edit-edit.svg"))
                    matlabel = ""
                    if mat:
                        matobj = FreeCAD.ActiveDocument.getObject(mat)
                        if matobj:
                            matlabel = matobj.Label
                    it3 = QtGui.QStandardItem(matlabel)
                    it3.setToolTip(mat)
                    omat = ""
                    if hasattr(obj, "Material") and obj.Material:
                        omat = obj.Material.Name
                        if omat != mat:
                            it3.setIcon(QtGui.QIcon(":/icons/edit-edit.svg"))
                    top.appendRow([it1, it2, it3])
            top.sortChildren(0)
        self.form.tree.expandAll()
        self.spanTopLevels()

    def updateByMaterial(self):
        groups = {}
        for name, rolemat in self.objectslist.items():
            role = rolemat[0]
            mat = rolemat[1]
            if not mat:
                mat = "Undefined"
            obj = FreeCAD.ActiveDocument.getObject(name)
            if obj:
                if (not self.form.onlyVisible.isChecked()) or obj.ViewObject.isVisible():
                    groups.setdefault(mat, []).append([name, role])

        for group in groups.keys():
            grlabel = "Undefined"
            if group != "Undefined":
                matobj = FreeCAD.ActiveDocument.getObject(group)
                if matobj:
                    grlabel = matobj.Label
            s1 = grlabel + " (" + str(len(groups[group])) + ")"
            top = QtGui.QStandardItem(s1)
            self.model.appendRow([top, QtGui.QStandardItem(), QtGui.QStandardItem()])
            for name, role in groups[group]:
                obj = FreeCAD.ActiveDocument.getObject(name)
                if obj:
                    it1 = QtGui.QStandardItem(obj.Label)
                    it1.setIcon(getIcon(obj))
                    it1.setToolTip(obj.Name)
                    it2 = QtGui.QStandardItem(role)
                    if role != self.getRole(obj):
                        it2.setIcon(QtGui.QIcon(":/icons/edit-edit.svg"))
                    mat = ""
                    matlabel = ""
                    if group != "Undefined":
                        matobj = FreeCAD.ActiveDocument.getObject(group)
                        if matobj:
                            matlabel = matobj.Label
                            mat = matobj.Name
                    it3 = QtGui.QStandardItem(matlabel)
                    it3.setToolTip(mat)
                    omat = ""
                    if hasattr(obj, "Material") and obj.Material:
                        omat = obj.Material.Name
                        if omat != mat:
                            it3.setIcon(QtGui.QIcon(":/icons/edit-edit.svg"))
                    top.appendRow([it1, it2, it3])
            top.sortChildren(0)
        self.form.tree.expandAll()
        self.spanTopLevels()

    def updateByTree(self):
        # order by hierarchy
        def istop(obj):
            for parent in obj.InListRecursive:
                if parent.Name in self.objectslist.keys():
                    return False
            return True

        rel = []
        deps = []
        for name in self.objectslist.keys():
            obj = FreeCAD.ActiveDocument.getObject(name)
            if obj:
                if istop(obj):
                    rel.append(obj)
                else:
                    deps.append(obj)
        pa = 1
        while deps:
            for obj in rel:
                for child in obj.OutList:
                    if child in deps:
                        rel.append(child)
                        deps.remove(child)
            pa += 1
            if pa == 10:  # max 10 hierarchy levels, okay? Let's keep civilised
                rel.extend(deps)
                break

        # print "rel:",[o.Label for o in rel]

        done = {}
        for obj in rel:
            rolemat = self.objectslist[obj.Name]
            role = rolemat[0]
            mat = rolemat[1]

            if (not self.form.onlyVisible.isChecked()) or obj.ViewObject.isVisible():
                it1 = QtGui.QStandardItem(obj.Label)
                it1.setIcon(getIcon(obj))
                it1.setToolTip(obj.Name)
                it2 = QtGui.QStandardItem(role)
                if role != self.getRole(obj):
                    it2.setIcon(QtGui.QIcon(":/icons/edit-edit.svg"))
                matlabel = ""
                if mat:
                    matobj = FreeCAD.ActiveDocument.getObject(mat)
                    if matobj:
                        matlabel = matobj.Label
                else:
                    mat = ""
                it3 = QtGui.QStandardItem(matlabel)
                it3.setToolTip(mat)
                omat = ""
                if hasattr(obj, "Material") and obj.Material:
                    omat = obj.Material.Name
                    if omat != mat:
                        it3.setIcon(QtGui.QIcon(":/icons/edit-edit.svg"))
                ok = False
                for par in obj.InListRecursive:
                    if par.Name in done:
                        if (not hasattr(par, "Hosts")) or (obj not in par.Hosts):
                            done[par.Name].appendRow([it1, it2, it3])
                            done[obj.Name] = it1
                            ok = True
                            break
                if not ok:
                    self.model.appendRow([it1, it2, it3])
                    done[obj.Name] = it1
        self.form.tree.expandAll()

    def updateDefault(self):
        for name, rolemat in self.objectslist.items():
            role = rolemat[0]
            mat = rolemat[1]
            obj = FreeCAD.ActiveDocument.getObject(name)
            if obj:
                if (not self.form.onlyVisible.isChecked()) or obj.ViewObject.isVisible():
                    it1 = QtGui.QStandardItem(obj.Label)
                    it1.setIcon(getIcon(obj))
                    it1.setToolTip(obj.Name)
                    it2 = QtGui.QStandardItem(role)
                    if role != self.getRole(obj):
                        it2.setIcon(QtGui.QIcon(":/icons/edit-edit.svg"))
                    matlabel = ""
                    if mat:
                        matobj = FreeCAD.ActiveDocument.getObject(mat)
                        if matobj:
                            matlabel = matobj.Label
                    else:
                        mat = ""
                    it3 = QtGui.QStandardItem(matlabel)
                    it3.setToolTip(mat)
                    omat = ""
                    if hasattr(obj, "Material") and obj.Material:
                        omat = obj.Material.Name
                        if omat != mat:
                            it3.setIcon(QtGui.QIcon(":/icons/edit-edit.svg"))
                    self.model.appendRow([it1, it2, it3])

    def spanTopLevels(self):
        if self.form.groupMode.currentIndex() in [1, 2]:
            idx = self.model.invisibleRootItem().index()
            for i in range(self.model.rowCount()):
                if self.model.item(i, 0).hasChildren():
                    self.form.tree.setFirstColumnSpanned(i, idx, True)

    def getRole(self, obj):
        if hasattr(obj, "IfcType"):
            return obj.IfcType
        elif hasattr(obj, "IfcRole"):
            return obj.IfcRole
        else:
            return None

    def onClickTree(self, index=None):
        FreeCADGui.Selection.clearSelection()
        sel = self.form.tree.selectedIndexes()
        mode = None
        mat = None
        for index in sel:
            if index.column() == 0:
                obj = FreeCAD.ActiveDocument.getObject(self.model.itemFromIndex(index).toolTip())
                if obj:
                    FreeCADGui.Selection.addSelection(obj)

        for index in sel:
            if index.column() == 1:
                if index.data() in self.ifctypes:
                    if mode:
                        if index.data() != mode:
                            mode = None
                            break
                    else:
                        mode = index.data()
        for index in sel:
            if index.column() == 2:
                item = self.model.itemFromIndex(index)
                m = FreeCAD.ActiveDocument.getObject(item.toolTip())
                if mat:
                    if m != mat:
                        mat = None
                        break
                else:
                    mat = m
        if mode:
            self.form.globalMode.setCurrentIndex(self.ifctypes.index(mode) + 1)
        else:
            self.form.globalMode.setCurrentIndex(0)
        if mat:
            self.form.globalMaterial.setCurrentIndex(self.materials.index(mat.Name) + 3)
        else:
            self.form.globalMaterial.setCurrentIndex(0)

    def onObjectTypeChanged(self, index=-1):
        changed = False
        if index >= 1:
            role = self.ifctypes[index - 1]
            sel = self.form.tree.selectedIndexes()
            for index in sel:
                if index.column() == 1:
                    if role:
                        if index.data() != role:
                            self.model.setData(index, role)
                            changed = True
        if changed:
            self.update()

    def onMaterialChanged(self, index=-1):
        changed = False
        if index == 1:
            FreeCADGui.runCommand("Arch_Material")
            QtCore.QTimer.singleShot(1000, self.checkMatChanged)
        elif index == 2:
            FreeCADGui.runCommand("Arch_MultiMaterial")
            QtCore.QTimer.singleShot(1000, self.checkMatChanged)
        elif index >= 3:
            mat = self.materials[index - 3]
            sel = self.form.tree.selectedIndexes()
            for index in sel:
                if index.column() == 2:
                    if mat:
                        mobj = FreeCAD.ActiveDocument.getObject(mat)
                        if mobj:
                            item = self.model.itemFromIndex(index)
                            if item.toolTip() != mat:
                                item.setText(mobj.Label)
                                item.setToolTip(mat)
                                changed = True
        if changed:
            self.update()

    def checkMatChanged(self):
        import Draft
        from PySide import QtCore, QtGui

        if getattr(self, "form", None) is None:
            return
        if FreeCADGui.Control.activeDialog():
            QtCore.QTimer.singleShot(500, self.checkMatChanged)
            return
        mats = [
            o.Name
            for o in FreeCAD.ActiveDocument.Objects
            if (o.isDerivedFrom("App::MaterialObject") or (Draft.getType(o) == "MultiMaterial"))
        ]
        if len(mats) != len(self.materials):
            newmats = [m for m in mats if not m in self.materials]
            self.materials = mats
            self.form.globalMaterial.clear()
            self.form.globalMaterial.addItem(" ")
            self.form.globalMaterial.addItem(translate("BIM", "Create new material"))
            self.form.globalMaterial.addItem(translate("BIM", "Create new multi-material"))
            for m in self.materials:
                o = FreeCAD.ActiveDocument.getObject(m)
                if o:
                    self.form.globalMaterial.addItem(
                        o.Label, QtGui.QIcon(":/icons/Arch_Material.svg")
                    )
            changed = False
            sel = self.form.tree.selectedIndexes()
            for index in sel:
                if index.column() == 2:
                    for mat in newmats:
                        mobj = FreeCAD.ActiveDocument.getObject(mat)
                        if mobj:
                            item = self.model.itemFromIndex(index)
                            if item.toolTip() != mat:
                                item.setText(mobj.Label)
                                item.setToolTip(mat)
                                changed = True
            if changed:
                self.update()

    def accept(self):
        # get current state of tree

        self.form.hide()
        for row in range(self.model.rowCount()):
            name = self.model.item(row, 0).toolTip()
            mat = self.model.item(row, 2).toolTip()
            if name:
                self.objectslist[name] = [self.model.item(row, 1).text(), mat]
            if self.model.item(row, 0).hasChildren():
                for childrow in range(self.model.item(row, 0).rowCount()):
                    name = self.model.item(row, 0).child(childrow, 0).toolTip()
                    mat = self.model.item(row, 0).child(childrow, 2).toolTip()
                    if name:
                        self.objectslist[name] = [
                            self.model.item(row, 0).child(childrow, 1).text(),
                            mat,
                        ]
        changed = False
        for name, rolemat in self.objectslist.items():
            role = rolemat[0]
            mat = rolemat[1]
            obj = FreeCAD.ActiveDocument.getObject(name)
            if obj:
                if hasattr(obj, "IfcRole") and (obj.IfcRole != role):
                    if not changed:
                        FreeCAD.ActiveDocument.openTransaction("Change IFC role")
                        changed = True
                    obj.IfcRole = role
                elif hasattr(obj, "IfcType") and (obj.IfcType != role):
                    if not changed:
                        FreeCAD.ActiveDocument.openTransaction("Change IFC type")
                        changed = True
                    obj.IfcType = role
                if mat and hasattr(obj, "Material"):
                    mobj = FreeCAD.ActiveDocument.getObject(mat)
                    if mobj:
                        if obj.Material:
                            if obj.Material.Name != mat:
                                if not changed:
                                    FreeCAD.ActiveDocument.openTransaction("Change material")
                                    changed = True
                                obj.Material = mobj
                        else:
                            if not changed:
                                FreeCAD.ActiveDocument.openTransaction("Change material")
                                changed = True
                            obj.Material = mobj
        if changed:
            FreeCAD.ActiveDocument.commitTransaction()
            FreeCAD.ActiveDocument.recompute()
        return self.reject()

    def reject(self):
        self.form.hide()
        del self.form
        return True


if FreeCAD.GuiUp:
    from PySide import QtCore, QtGui

    class IfcElementsDelegate(QtGui.QStyledItemDelegate):
        def __init__(self, parent=None, dialog=None, *args):
            import Arch_rc

            try:
                import ArchIFC

                self.roles = ArchIFC.IfcTypes
            except (ImportError, AttributeError):
                import ArchComponent

                self.roles = ArchComponent.IfcRoles
            self.mats = []
            self.matlabels = []
            for o in FreeCAD.ActiveDocument.Objects:
                if o.isDerivedFrom("App::MaterialObject"):
                    self.mats.append(o.Name)
                    self.matlabels.append(o.Label)
            self.dialog = dialog
            self.btn = QtGui.QPushButton()
            self.btn.setIcon(QtGui.QIcon(":/icons/IFC.svg"))
            self.btn.setText("")
            QtGui.QStyledItemDelegate.__init__(self, parent, *args)

        def paint(self, painter, option, index):
            # not used - ugly and fake!
            if index.column() == 3:
                self.btn.setGeometry(option.rect)
                if option.state == QtGui.QStyle.State_Selected:
                    painter.fillRect(option.rect, option.palette.highlight())
                p = QtGui.QPixmap.grabWidget(self.btn)
                painter.drawPixmap(option.rect.x(), option.rect.y(), p)
            else:
                QtGui.QStyledItemDelegate.paint(self, painter, option, index)

        def createEditor(self, parent, option, index):
            if index.column() > 0:
                editor = QtGui.QComboBox(parent)
            else:
                editor = QtGui.QLineEdit(parent)
            return editor

        def setEditorData(self, editor, index):
            if index.column() == 1:
                idx = -1
                editor.addItems(self.roles)
                if index.data() in self.roles:
                    idx = self.roles.index(index.data())
                editor.setCurrentIndex(idx)
            elif index.column() == 2:
                idx = -1
                editor.addItems(self.matlabels)
                item = index.model().itemFromIndex(index)
                if item.toolTip() in self.mats:
                    idx = self.mats.index(item.toolTip())
                editor.setCurrentIndex(idx)
            else:
                editor.setText(index.data())

        def setModelData(self, editor, model, index):
            if index.column() == 1:
                if editor.currentIndex() == -1:
                    model.setData(index, "")
                else:
                    model.setData(index, self.roles[editor.currentIndex()])
            elif index.column() == 2:
                if editor.currentIndex() > -1:
                    idx = editor.currentIndex()
                    item = model.itemFromIndex(index)
                    item.setText(self.matlabels[idx])
                    item.setToolTip(self.mats[idx])
            else:
                model.setData(index, editor.text())
                item = model.itemFromIndex(index)
                obj = FreeCAD.ActiveDocument.getObject(item.toolTip())
                if obj:
                    obj.Label = editor.text()
            self.dialog.update()


def getIcon(obj):
    """returns a QIcon for an object"""

    from PySide import QtGui
    import Arch_rc

    if hasattr(obj.ViewObject, "Icon"):
        return obj.ViewObject.Icon
    elif hasattr(obj.ViewObject, "Proxy") and hasattr(obj.ViewObject.Proxy, "getIcon"):
        icon = obj.ViewObject.Proxy.getIcon()
        if icon.startswith("/*"):
            return QtGui.QIcon(QtGui.QPixmap(icon))
        else:
            return QtGui.QIcon(icon)
    else:
        return QtGui.QIcon(":/icons/Arch_Component.svg")


FreeCADGui.addCommand("BIM_IfcElements", BIM_IfcElements())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2024 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

import FreeCAD


class CyclicSelectionObserver:
    def addSelection(self, document, object, element, position):
        import FreeCADGui

        if not FreeCAD.ActiveDocument:
            return
        if not hasattr(FreeCAD, "CyclicSelectionObserver"):
            return
        FreeCADGui.Selection.removeSelection(FreeCAD.ActiveDocument.getObject(object))
        FreeCADGui.Selection.removeObserver(FreeCAD.CyclicSelectionObserver)
        del FreeCAD.CyclicSelectionObserver
        preselection = FreeCADGui.Selection.getPreselection()
        FreeCADGui.Selection.addSelection(
            FreeCAD.ActiveDocument.getObject(preselection.Object.Name),
            preselection.SubElementNames[0],
        )
        FreeCAD.ActiveDocument.recompute()


class CyclicObjectSelector:
    def __init__(self):
        self.selectableObjects = []
        self.objectIndex = 0

    def selectObject(self, event_callback):
        import FreeCADGui
        from pivy import coin

        if not FreeCAD.ActiveDocument:
            return
        event = event_callback.getEvent()

        if event.getState() != coin.SoMouseButtonEvent.DOWN or not self.selectableObjects:
            return

        pos = event.getPosition().getValue()
        element_list = FreeCADGui.ActiveDocument.ActiveView.getObjectsInfo(
            (int(pos[0]), int(pos[1]))
        )

        if not element_list:
            self.selectableObjects = []
            if hasattr(FreeCAD, "CyclicSelectionObserver"):
                FreeCADGui.Selection.removeObserver(FreeCAD.CyclicSelectionObserver)
                del FreeCAD.CyclicSelectionObserver
            return

        FreeCAD.CyclicSelectionObserver = CyclicSelectionObserver()
        FreeCADGui.Selection.addObserver(FreeCAD.CyclicSelectionObserver)

    def cycleSelectableObjects(self, event_callback):
        import FreeCADGui

        if not FreeCAD.ActiveDocument:
            return
        event = event_callback.getEvent()

        if not event.isKeyPressEvent(event, event.TAB):
            return

        pos = event.getPosition().getValue()
        selectableObjects = FreeCADGui.ActiveDocument.ActiveView.getObjectsInfo(
            (int(pos[0]), int(pos[1]))
        )

        if not selectableObjects:
            return

        if self.selectableObjects != selectableObjects:
            self.selectableObjects = selectableObjects
            self.objectIndex = 0
        elif self.objectIndex < len(self.selectableObjects) - 1:
            self.objectIndex += 1
        else:
            self.objectIndex = 0
        object_name = self.selectableObjects[self.objectIndex]["Object"]
        subelement_name = self.selectableObjects[self.objectIndex]["Component"]
        FreeCADGui.getMainWindow().showMessage(
            "Cycle preselected (TAB): {} - {}".format(object_name, subelement_name), 0
        )
        FreeCADGui.Selection.setPreselection(
            FreeCAD.ActiveDocument.getObject(object_name), subelement_name
        )


class Setup:
    def slotActivateDocument(self, doc):
        from pivy import coin

        cos = CyclicObjectSelector()
        if doc and doc.ActiveView and hasattr(doc.ActiveView, "getSceneGraph"):
            self.callback = doc.ActiveView.addEventCallbackPivy(
                coin.SoMouseButtonEvent.getClassTypeId(), cos.selectObject
            )
            self.callback = doc.ActiveView.addEventCallbackPivy(
                coin.SoKeyboardEvent.getClassTypeId(), cos.cycleSelectableObjects
            )



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2017 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

from Path.Dressup.Gui.TagPreferences import HoldingTagPreferences
from PathScripts.PathUtils import waiting_effects
from PySide.QtCore import QT_TRANSLATE_NOOP
import FreeCAD
import Path
import Path.Dressup.Utils as PathDressup
import PathScripts.PathUtils as PathUtils
import copy
import math


# lazily loaded modules
from lazy_loader.lazy_loader import LazyLoader

Part = LazyLoader("Part", globals(), "Part")

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())

translate = FreeCAD.Qt.translate


def debugEdge(edge, prefix, force=False):
    if force or Path.Log.getLevel(Path.Log.thisModule()) == Path.Log.Level.DEBUG:
        pf = edge.valueAt(edge.FirstParameter)
        pl = edge.valueAt(edge.LastParameter)
        if type(edge.Curve) in [Part.Line, Part.LineSegment]:
            print(
                "%s %s((%.2f, %.2f, %.2f) - (%.2f, %.2f, %.2f))"
                % (prefix, type(edge.Curve), pf.x, pf.y, pf.z, pl.x, pl.y, pl.z)
            )
        else:
            pm = edge.valueAt((edge.FirstParameter + edge.LastParameter) / 2)
            print(
                "%s %s((%.2f, %.2f, %.2f) - (%.2f, %.2f, %.2f) - (%.2f, %.2f, %.2f))"
                % (
                    prefix,
                    type(edge.Curve),
                    pf.x,
                    pf.y,
                    pf.z,
                    pm.x,
                    pm.y,
                    pm.z,
                    pl.x,
                    pl.y,
                    pl.z,
                )
            )


def debugMarker(vector, label, color=None, radius=0.5):
    if Path.Log.getLevel(Path.Log.thisModule()) == Path.Log.Level.DEBUG:
        obj = FreeCAD.ActiveDocument.addObject("Part::Sphere", label)
        obj.Label = label
        obj.Radius = radius
        obj.Placement = FreeCAD.Placement(vector, FreeCAD.Rotation(FreeCAD.Vector(0, 0, 1), 0))
        if color:
            obj.ViewObject.ShapeColor = color


def debugCylinder(vector, r, height, label, color=None):
    if Path.Log.getLevel(Path.Log.thisModule()) == Path.Log.Level.DEBUG:
        obj = FreeCAD.ActiveDocument.addObject("Part::Cylinder", label)
        obj.Label = label
        obj.Radius = r
        obj.Height = height
        obj.Placement = FreeCAD.Placement(vector, FreeCAD.Rotation(FreeCAD.Vector(0, 0, 1), 0))
        obj.ViewObject.Transparency = 90
        if color:
            obj.ViewObject.ShapeColor = color


def debugCone(vector, r1, r2, height, label, color=None):
    if Path.Log.getLevel(Path.Log.thisModule()) == Path.Log.Level.DEBUG:
        obj = FreeCAD.ActiveDocument.addObject("Part::Cone", label)
        obj.Label = label
        obj.Radius1 = r1
        obj.Radius2 = r2
        obj.Height = height
        obj.Placement = FreeCAD.Placement(vector, FreeCAD.Rotation(FreeCAD.Vector(0, 0, 1), 0))
        obj.ViewObject.Transparency = 90
        if color:
            obj.ViewObject.ShapeColor = color


class Tag:
    def __init__(self, nr, x, y, width, height, angle, radius, enabled=True):
        Path.Log.track(
            "%.2f, %.2f, %.2f, %.2f, %.2f, %.2f, %d" % (x, y, width, height, angle, radius, enabled)
        )
        self.nr = nr
        self.x = x
        self.y = y
        self.width = math.fabs(width)
        self.height = math.fabs(height)
        self.actualHeight = self.height
        self.angle = math.fabs(angle)
        self.radius = getattr(
            radius, "Value", FreeCAD.Units.Quantity(radius, FreeCAD.Units.Length).Value
        )
        self.enabled = enabled
        self.isSquare = False

        # initialized later
        self.toolRadius = None
        self.realRadius = None
        self.r1 = None
        self.r2 = None
        self.solid = None
        self.z = None

    def fullWidth(self):
        return 2 * self.toolRadius + self.width

    def originAt(self, z):
        return FreeCAD.Vector(self.x, self.y, z)

    def bottom(self):
        return self.z

    def top(self):
        return self.z + self.actualHeight

    def createSolidsAt(self, z, R):
        self.z = z
        self.toolRadius = R
        r1 = self.fullWidth() / 2
        self.r1 = r1
        self.r2 = r1
        height = self.height * 1.01
        radius = 0
        if Path.Geom.isRoughly(90, self.angle) and height > 0:
            # cylinder
            self.isSquare = True
            self.solid = Part.makeCylinder(r1, height)
            radius = min(min(self.radius, r1), self.height)
            Path.Log.debug("Part.makeCylinder(%f, %f)" % (r1, height))
        elif self.angle > 0.0 and height > 0.0:
            # cone
            rad = math.radians(self.angle)
            tangens = math.tan(rad)
            dr = height / tangens
            if dr < r1:
                # with top
                r2 = r1 - dr
                s = height / math.sin(rad)
                radius = min(r2, s) * math.tan((math.pi - rad) / 2) * 0.95
            else:
                # triangular
                r2 = 0
                height = r1 * tangens * 1.01
                self.actualHeight = height
            self.r2 = r2
            Path.Log.debug("Part.makeCone(%f, %f, %f)" % (r1, r2, height))
            self.solid = Part.makeCone(r1, r2, height)
        else:
            # degenerated case - no tag
            Path.Log.debug("Part.makeSphere(%f / 10000)" % (r1))
            self.solid = Part.makeSphere(r1 / 10000)
        if not Path.Geom.isRoughly(0, R):  # testing is easier if the solid is not rotated
            angle = -Path.Geom.getAngle(self.originAt(0)) * 180 / math.pi
            Path.Log.debug("solid.rotate(%f)" % angle)
            self.solid.rotate(FreeCAD.Vector(0, 0, 0), FreeCAD.Vector(0, 0, 1), angle)
        orig = self.originAt(z - 0.01 * self.actualHeight)
        Path.Log.debug("solid.translate(%s)" % orig)
        self.solid.translate(orig)
        radius = min(self.radius, radius)
        self.realRadius = radius
        if not Path.Geom.isRoughly(0, radius):
            Path.Log.debug("makeFillet(%.4f)" % radius)
            self.solid = self.solid.makeFillet(radius, [self.solid.Edges[0]])

    def filterIntersections(self, pts, face):
        if type(face.Surface) in [Part.Cone, Part.Cylinder, Part.Toroid]:
            Path.Log.track("it's a cone/cylinder, checking z")
            return list([pt for pt in pts if pt.z >= self.bottom() and pt.z <= self.top()])
        if type(face.Surface) is Part.Plane:
            Path.Log.track("it's a plane, checking R")
            c = face.Edges[0].Curve
            if type(c) is Part.Circle:
                return list(
                    [
                        pt
                        for pt in pts
                        if (pt - c.Center).Length <= c.Radius
                        or Path.Geom.isRoughly((pt - c.Center).Length, c.Radius)
                    ]
                )
        Path.Log.error("==== we got a %s" % face.Surface)

    def isPointOnEdge(self, pt, edge):
        param = edge.Curve.parameter(pt)
        if edge.FirstParameter <= param <= edge.LastParameter:
            return True
        if edge.LastParameter <= param <= edge.FirstParameter:
            return True
        if Path.Geom.isRoughly(edge.FirstParameter, param) or Path.Geom.isRoughly(
            edge.LastParameter, param
        ):
            return True
        # print("-------- X %.2f <= %.2f <=%.2f   (%.2f, %.2f, %.2f)   %.2f:%.2f" % (edge.FirstParameter, param, edge.LastParameter, pt.x, pt.y, pt.z, edge.Curve.parameter(edge.valueAt(edge.FirstParameter)), edge.Curve.parameter(edge.valueAt(edge.LastParameter))))
        # p1 = edge.Vertexes[0]
        # f1 = edge.Curve.parameter(FreeCAD.Vector(p1.X, p1.Y, p1.Z))
        # p2 = edge.Vertexes[1]
        # f2 = edge.Curve.parameter(FreeCAD.Vector(p2.X, p2.Y, p2.Z))
        return False

    def nextIntersectionClosestTo(self, edge, solid, refPt):
        # debugEdge(edge, 'intersects_')

        vertexes = edge.common(solid).Vertexes
        if vertexes:
            pt = sorted(vertexes, key=lambda v: (v.Point - refPt).Length)[0].Point
            debugEdge(
                edge,
                "intersects (%.2f, %.2f, %.2f) -> (%.2f, %.2f, %.2f)"
                % (refPt.x, refPt.y, refPt.z, pt.x, pt.y, pt.z),
            )
            return pt
        return None

    def intersects(self, edge, param):
        def isDefinitelySmaller(z, zRef):
            # Eliminate false positives of edges that just brush along the top of the tag
            return z < zRef and not Path.Geom.isRoughly(z, zRef, 0.01)

        if self.enabled:
            zFirst = edge.valueAt(edge.FirstParameter).z
            zLast = edge.valueAt(edge.LastParameter).z
            zMax = self.top()
            if isDefinitelySmaller(zFirst, zMax) or isDefinitelySmaller(zLast, zMax):
                return self.nextIntersectionClosestTo(edge, self.solid, edge.valueAt(param))
        return None

    def bbEdges(self):
        edges = []
        for i in range(12):
            p1, p2 = self.solid.BoundBox.getEdge(i)
            edges.append(Part.Edge(Part.LineSegment(p1, p2)))
        return edges

    def bbShow(self):
        for e in self.bbEdges():
            Part.show(e)


class MapWireToTag:
    def __init__(self, edge, tag, i, segm, maxZ, hSpeed, vSpeed):
        debugEdge(edge, "MapWireToTag(%.2f, %.2f, %.2f)" % (i.x, i.y, i.z))
        self.tag = tag
        self.segm = segm
        self.maxZ = maxZ
        self.hSpeed = hSpeed
        self.vSpeed = vSpeed
        if Path.Geom.pointsCoincide(edge.valueAt(edge.FirstParameter), i):
            tail = edge
            self.commands = []
            debugEdge(tail, ".........=")
        elif Path.Geom.pointsCoincide(edge.valueAt(edge.LastParameter), i):
            debugEdge(edge, "++++++++ .")
            self.commands = Path.Geom.cmdsForEdge(
                edge, segm=segm, hSpeed=self.hSpeed, vSpeed=self.vSpeed
            )
            tail = None
        else:
            e, tail = Path.Geom.splitEdgeAt(edge, i)
            debugEdge(e, "++++++++ .")
            self.commands = Path.Geom.cmdsForEdge(
                e, segm=segm, hSpeed=self.hSpeed, vSpeed=self.vSpeed
            )
            debugEdge(tail, ".........-")
            self.initialEdge = edge
        self.tail = tail
        self.edges = []
        self.entry = i
        if tail:
            Path.Log.debug("MapWireToTag(%s - %s)" % (i, tail.valueAt(tail.FirstParameter)))
        else:
            Path.Log.debug("MapWireToTag(%s - )" % i)
        self.complete = False
        self.haveProblem = False

        # initialized later
        self.edgePoints = None
        self.edgesCleanup = None
        self.edgesOrder = None
        self.entryEdges = None
        self.exit = None
        self.exitEdges = None
        self.finalEdge = None
        self.offendingEdge = None
        self.realEntry = None
        self.realExit = None

    def addEdge(self, edge):
        debugEdge(edge, "..........")
        self.edges.append(edge)

    def needToFlipEdge(self, edge, p):
        if Path.Geom.pointsCoincide(edge.valueAt(edge.LastParameter), p):
            return True, edge.valueAt(edge.FirstParameter)
        return False, edge.valueAt(edge.LastParameter)

    def isEntryOrExitStrut(self, e):
        p1 = e.valueAt(e.FirstParameter)
        p2 = e.valueAt(e.LastParameter)
        if Path.Geom.pointsCoincide(p1, self.entry) and p2.z >= self.entry.z:
            return 1
        if Path.Geom.pointsCoincide(p2, self.entry) and p1.z >= self.entry.z:
            return 1
        if Path.Geom.pointsCoincide(p1, self.exit) and p2.z >= self.exit.z:
            return 2
        if Path.Geom.pointsCoincide(p2, self.exit) and p1.z >= self.exit.z:
            return 2
        return 0

    def cleanupEdges(self, edges):
        # want to remove all edges from the wire itself, and all internal struts
        Path.Log.track("+cleanupEdges")
        Path.Log.debug(" edges:")
        if not edges:
            return edges
        for e in edges:
            debugEdge(e, "   ")
        Path.Log.debug(":")
        self.edgesCleanup = [copy.copy(edges)]

        # remove any edge that has a point inside the tag solid
        # and collect all edges that are connected to the entry and/or exit
        self.entryEdges = []
        self.exitEdges = []
        self.edgePoints = []
        for e in copy.copy(edges):
            p1 = e.valueAt(e.FirstParameter)
            p2 = e.valueAt(e.LastParameter)
            self.edgePoints.append(p1)
            self.edgePoints.append(p2)
            if self.tag.solid.isInside(p1, Path.Geom.Tolerance, False) or self.tag.solid.isInside(
                p2, Path.Geom.Tolerance, False
            ):
                edges.remove(e)
                debugEdge(e, "......... X0", False)
            else:
                if Path.Geom.pointsCoincide(p1, self.entry) or Path.Geom.pointsCoincide(
                    p2, self.entry
                ):
                    self.entryEdges.append(e)
                if Path.Geom.pointsCoincide(p1, self.exit) or Path.Geom.pointsCoincide(
                    p2, self.exit
                ):
                    self.exitEdges.append(e)
        self.edgesCleanup.append(copy.copy(edges))

        # if there are no edges connected to entry/exit, it means the plunge in/out is vertical
        # we need to add in the missing segment and collect the new entry/exit edges.
        if not self.entryEdges:
            Path.Log.debug("fill entryEdges…")
            self.realEntry = sorted(self.edgePoints, key=lambda p: (p - self.entry).Length)[0]
            self.entryEdges = list(
                [e for e in edges if Path.Geom.edgeConnectsTo(e, self.realEntry)]
            )
            edges.append(Part.Edge(Part.LineSegment(self.entry, self.realEntry)))
        else:
            self.realEntry = None
        if not self.exitEdges:
            Path.Log.debug("fill exitEdges…")
            self.realExit = sorted(self.edgePoints, key=lambda p: (p - self.exit).Length)[0]
            self.exitEdges = list([e for e in edges if Path.Geom.edgeConnectsTo(e, self.realExit)])
            edges.append(Part.Edge(Part.LineSegment(self.realExit, self.exit)))
        else:
            self.realExit = None
        self.edgesCleanup.append(copy.copy(edges))

        # if there are 2 edges attached to entry/exit, throw away the one that is "lower"
        if len(self.entryEdges) > 1:
            debugEdge(self.entryEdges[0], " entry[0]", False)
            debugEdge(self.entryEdges[1], " entry[1]", False)
            if self.entryEdges[0].BoundBox.ZMax < self.entryEdges[1].BoundBox.ZMax:
                edges.remove(self.entryEdges[0])
                debugEdge(e, "......... X1", False)
            else:
                edges.remove(self.entryEdges[1])
                debugEdge(e, "......... X2", False)
        if len(self.exitEdges) > 1:
            debugEdge(self.exitEdges[0], " exit[0]", False)
            debugEdge(self.exitEdges[1], " exit[1]", False)
            if self.exitEdges[0].BoundBox.ZMax < self.exitEdges[1].BoundBox.ZMax:
                if self.exitEdges[0] in edges:
                    edges.remove(self.exitEdges[0])
                debugEdge(e, "......... X3", False)
            else:
                if self.exitEdges[1] in edges:
                    edges.remove(self.exitEdges[1])
                debugEdge(e, "......... X4", False)

        self.edgesCleanup.append(copy.copy(edges))
        return edges

    def orderAndFlipEdges(self, inputEdges):
        Path.Log.track(
            "entry(%.2f, %.2f, %.2f), exit(%.2f, %.2f, %.2f)"
            % (
                self.entry.x,
                self.entry.y,
                self.entry.z,
                self.exit.x,
                self.exit.y,
                self.exit.z,
            )
        )
        self.edgesOrder = []
        outputEdges = []
        p0 = self.entry
        lastP = p0
        edges = copy.copy(inputEdges)
        while edges:
            # print("(%.2f, %.2f, %.2f) %d %d" % (p0.x, p0.y, p0.z))
            for e in copy.copy(edges):
                p1 = e.valueAt(e.FirstParameter)
                p2 = e.valueAt(e.LastParameter)
                if Path.Geom.pointsCoincide(p1, p0):
                    outputEdges.append((e, False))
                    edges.remove(e)
                    lastP = None
                    p0 = p2
                    debugEdge(e, ">>>>> no flip")
                    break
                elif Path.Geom.pointsCoincide(p2, p0):
                    flipped = Path.Geom.flipEdge(e)
                    if flipped is not None:
                        outputEdges.append((flipped, True))
                    else:
                        p0 = None
                        cnt = 0
                        for p in reversed(e.discretize(Deflection=0.01)):
                            if p0 is not None:
                                outputEdges.append((Part.Edge(Part.LineSegment(p0, p)), True))
                                cnt = cnt + 1
                            p0 = p
                        Path.Log.info("replaced edge with %d straight segments" % cnt)
                    edges.remove(e)
                    lastP = None
                    p0 = p1
                    debugEdge(e, ">>>>> flip")
                    break
                else:
                    debugEdge(e, "<<<<< (%.2f, %.2f, %.2f)" % (p0.x, p0.y, p0.z))

            if lastP == p0:
                self.edgesOrder.append(outputEdges)
                self.edgesOrder.append(edges)
                Path.Log.debug("input edges:")
                for e in inputEdges:
                    debugEdge(e, "  ", False)
                Path.Log.debug("ordered edges:")
                for e, flip in outputEdges:
                    debugEdge(e, "  %c " % ("<" if flip else ">"), False)
                Path.Log.debug("remaining edges:")
                for e in edges:
                    debugEdge(e, "    ", False)
                raise ValueError("No connection to %s" % (p0))
            elif lastP:
                Path.Log.debug(
                    "xxxxxx (%.2f, %.2f, %.2f) (%.2f, %.2f, %.2f)"
                    % (p0.x, p0.y, p0.z, lastP.x, lastP.y, lastP.z)
                )
            else:
                Path.Log.debug("xxxxxx (%.2f, %.2f, %.2f) -" % (p0.x, p0.y, p0.z))
            lastP = p0
        Path.Log.track("-")
        return outputEdges

    def isStrut(self, edge):
        p1 = Path.Geom.xy(edge.valueAt(edge.FirstParameter))
        p2 = Path.Geom.xy(edge.valueAt(edge.LastParameter))
        return Path.Geom.pointsCoincide(p1, p2)

    def shell(self):
        if len(self.edges) > 1:
            wire = Part.Wire(self.initialEdge)
        else:
            edge = self.edges[0]
            if Path.Geom.pointsCoincide(
                edge.valueAt(edge.FirstParameter),
                self.finalEdge.valueAt(self.finalEdge.FirstParameter),
            ):
                wire = Part.Wire(self.finalEdge)
            elif hasattr(self, "initialEdge") and Path.Geom.pointsCoincide(
                edge.valueAt(edge.FirstParameter),
                self.initialEdge.valueAt(self.initialEdge.FirstParameter),
            ):
                wire = Part.Wire(self.initialEdge)
            else:
                wire = Part.Wire(edge)

        for edge in self.edges[1:]:
            if Path.Geom.pointsCoincide(
                edge.valueAt(edge.FirstParameter),
                self.finalEdge.valueAt(self.finalEdge.FirstParameter),
            ):
                wire.add(self.finalEdge)
            else:
                wire.add(edge)

        shell = wire.extrude(FreeCAD.Vector(0, 0, self.tag.height + 1))
        nullFaces = list([f for f in shell.Faces if Path.Geom.isRoughly(f.Area, 0)])
        if nullFaces:
            return shell.removeShape(nullFaces)
        return shell

    def commandsForEdges(self):
        if self.edges:
            try:
                shape = self.shell().common(self.tag.solid)
                commands = []
                rapid = None
                for e, flip in self.orderAndFlipEdges(self.cleanupEdges(shape.Edges)):
                    debugEdge(e, "++++++++ %s" % ("<" if flip else ">"), False)
                    p1 = e.valueAt(e.FirstParameter)
                    p2 = e.valueAt(e.LastParameter)
                    if (
                        self.tag.isSquare
                        and (Path.Geom.isRoughly(p1.z, self.maxZ) or p1.z > self.maxZ)
                        and (Path.Geom.isRoughly(p2.z, self.maxZ) or p2.z > self.maxZ)
                    ):
                        rapid = p1 if flip else p2
                    else:
                        if rapid:
                            commands.append(
                                Path.Command("G0", {"X": rapid.x, "Y": rapid.y, "Z": rapid.z})
                            )
                            rapid = None
                        commands.extend(
                            Path.Geom.cmdsForEdge(
                                e,
                                False,
                                False,
                                self.segm,
                                hSpeed=self.hSpeed,
                                vSpeed=self.vSpeed,
                            )
                        )
                if rapid:
                    commands.append(Path.Command("G0", {"X": rapid.x, "Y": rapid.y, "Z": rapid.z}))
                    # rapid = None  # commented out per LGTM suggestion
                return commands
            except Exception as e:
                Path.Log.error(
                    "Exception during processing tag @(%.2f, %.2f) (%s) - disabling the tag"
                    % (self.tag.x, self.tag.y, e.args[0])
                )
                self.tag.enabled = False
                commands = []
                for e in self.edges:
                    commands.extend(
                        Path.Geom.cmdsForEdge(e, hSpeed=self.hSpeed, vSpeed=self.vSpeed)
                    )
                return commands
        return []

    def add(self, edge):
        self.tail = None
        self.finalEdge = edge
        if self.tag.solid.isInside(edge.valueAt(edge.LastParameter), Path.Geom.Tolerance, True):
            Path.Log.track("solid.isInside")
            self.addEdge(edge)
        else:
            i = self.tag.intersects(edge, edge.LastParameter)
            if not i:
                self.offendingEdge = edge
                debugEdge(edge, "offending Edge:", False)
                o = self.tag.originAt(self.tag.z)
                Path.Log.debug("originAt: (%.2f, %.2f, %.2f)" % (o.x, o.y, o.z))
                i = edge.valueAt(edge.FirstParameter)
            if Path.Geom.pointsCoincide(i, edge.valueAt(edge.FirstParameter)):
                Path.Log.track("tail")
                self.tail = edge
            else:
                Path.Log.track("split")
                e, tail = Path.Geom.splitEdgeAt(edge, i)
                self.addEdge(e)
                self.tail = tail
            self.exit = i
            self.complete = True
            self.commands.extend(self.commandsForEdges())

    def mappingComplete(self):
        return self.complete


class _RapidEdges:
    def __init__(self, rapid):
        self.rapid = rapid

    def isRapid(self, edge):
        if type(edge.Curve) in [Part.Line, Part.LineSegment]:
            v0 = edge.Vertexes[0]
            v1 = edge.Vertexes[1]
            for r in self.rapid:
                r0 = r.Vertexes[0]
                r1 = r.Vertexes[1]
                if (
                    Path.Geom.isRoughly(r0.X, v0.X)
                    and Path.Geom.isRoughly(r0.Y, v0.Y)
                    and Path.Geom.isRoughly(r0.Z, v0.Z)
                    and Path.Geom.isRoughly(r1.X, v1.X)
                    and Path.Geom.isRoughly(r1.Y, v1.Y)
                    and Path.Geom.isRoughly(r1.Z, v1.Z)
                ):
                    return True
        return False


class PathData:
    def __init__(self, obj):
        Path.Log.track(obj.Base.Name)
        self.obj = obj
        path = PathUtils.getPathWithPlacement(obj.Base)
        self.wire, rapid, rapid_indexes = Path.Geom.wireForPath(path)
        self.rapid = _RapidEdges(rapid)
        if self.wire:
            self.edges = self.wire.Edges
        else:
            self.edges = []
        self.baseWire = self.findBottomWire(self.edges)

    def findBottomWire(self, edges):
        (minZ, maxZ) = self.findZLimits(edges)
        self.minZ = minZ
        self.maxZ = maxZ
        bottom = [
            e
            for e in edges
            if Path.Geom.isRoughly(e.Vertexes[0].Point.z, minZ)
            and Path.Geom.isRoughly(e.Vertexes[1].Point.z, minZ)
        ]
        self.bottomEdges = bottom
        try:
            wire = Part.Wire(bottom)
            if wire.isClosed():
                return wire
        except Exception:
            return None

    def supportsTagGeneration(self):
        return self.baseWire is not None

    def findZLimits(self, edges):
        # not considering arcs and spheres in Z direction, find the highest and lowest Z values
        minZ = 99999999999
        maxZ = -99999999999
        for e in edges:
            if self.rapid.isRapid(e):
                continue
            for v in e.Vertexes:
                if v.Point.z < minZ:
                    minZ = v.Point.z
                if v.Point.z > maxZ:
                    maxZ = v.Point.z
        return (minZ, maxZ)

    def shortestAndLongestPathEdge(self):
        edges = sorted(self.bottomEdges, key=lambda e: e.Length)
        return (edges[0], edges[-1])

    def generateTags(
        self, obj, count, width=None, height=None, angle=None, radius=None, spacing=None
    ):
        Path.Log.track(count, width, height, angle, spacing)
        # for e in self.baseWire.Edges:
        #    debugMarker(e.Vertexes[0].Point, 'base', (0.0, 1.0, 1.0), 0.2)

        if spacing:
            tagDistance = spacing
        else:
            tagDistance = self.baseWire.Length / (count if count else 4)

        W = width if width else self.defaultTagWidth()
        H = height if height else self.defaultTagHeight()
        A = angle if angle else self.defaultTagAngle()
        R = radius if radius else self.defaultTagRadius()

        # start assigning tags on the longest segment
        (shortestEdge, longestEdge) = self.shortestAndLongestPathEdge()
        startIndex = 0
        for i in range(0, len(self.baseWire.Edges)):
            edge = self.baseWire.Edges[i]
            Path.Log.debug("  %d: %.2f" % (i, edge.Length))
            if Path.Geom.isRoughly(edge.Length, longestEdge.Length):
                startIndex = i
                break

        startEdge = self.baseWire.Edges[startIndex]
        startCount = int(startEdge.Length / tagDistance)
        if (longestEdge.Length - shortestEdge.Length) > shortestEdge.Length:
            startCount = int(startEdge.Length / tagDistance) + 1

        lastTagLength = (startEdge.Length + (startCount - 1) * tagDistance) / 2
        currentLength = startEdge.Length

        minLength = min(2.0 * W, longestEdge.Length)

        Path.Log.debug(
            "length=%.2f shortestEdge=%.2f(%.2f) longestEdge=%.2f(%.2f) minLength=%.2f"
            % (
                self.baseWire.Length,
                shortestEdge.Length,
                shortestEdge.Length / self.baseWire.Length,
                longestEdge.Length,
                longestEdge.Length / self.baseWire.Length,
                minLength,
            )
        )
        Path.Log.debug(
            "   start: index=%-2d count=%d (length=%.2f, distance=%.2f)"
            % (startIndex, startCount, startEdge.Length, tagDistance)
        )
        Path.Log.debug("               -> lastTagLength=%.2f)" % lastTagLength)
        Path.Log.debug("               -> currentLength=%.2f)" % currentLength)

        edgeDict = {startIndex: startCount}

        for i in range(startIndex + 1, len(self.baseWire.Edges)):
            edge = self.baseWire.Edges[i]
            (currentLength, lastTagLength) = self.processEdge(
                i, edge, currentLength, lastTagLength, tagDistance, minLength, edgeDict
            )
        for i in range(0, startIndex):
            edge = self.baseWire.Edges[i]
            (currentLength, lastTagLength) = self.processEdge(
                i, edge, currentLength, lastTagLength, tagDistance, minLength, edgeDict
            )

        tags = []

        for i, count in edgeDict.items():
            edge = self.baseWire.Edges[i]
            Path.Log.debug(" %d: %d" % (i, count))
            # debugMarker(edge.Vertexes[0].Point, 'base', (1.0, 0.0, 0.0), 0.2)
            # debugMarker(edge.Vertexes[1].Point, 'base', (0.0, 1.0, 0.0), 0.2)
            if 0 != count:
                distance = (edge.LastParameter - edge.FirstParameter) / count
                for j in range(0, count):
                    tag = edge.Curve.value((j + 0.5) * distance)
                    tags.append(Tag(j, tag.x, tag.y, W, H, A, R, True))

        return tags

    def copyTags(self, obj, fromObj, width, height, angle, radius, production=True):
        print(
            "copyTags(%s, %s, %.2f, %.2f, %.2f, %.2f"
            % (obj.Label, fromObj.Label, width, height, angle, radius)
        )
        W = width if width else self.defaultTagWidth()
        H = height if height else self.defaultTagHeight()
        A = angle if angle else self.defaultTagAngle()
        R = radius if radius else self.defaultTagRadius()

        tags = []
        j = 0
        for i, pos in enumerate(fromObj.Positions):
            print("tag[%d]" % i)
            if i not in fromObj.Disabled:
                dist = self.baseWire.distToShape(
                    Part.Vertex(FreeCAD.Vector(pos.x, pos.y, self.minZ))
                )
                if production or dist[0] < W:
                    # russ4262:: `production` variable was a `True` declaration, forcing True branch to be processed always
                    #   The application of the `production` argument/variable is to appease LGTM
                    print("tag[%d/%d]: (%.2f, %.2f, %.2f)" % (i, j, pos.x, pos.y, self.minZ))
                    at = dist[1][0][0]
                    tags.append(Tag(j, at.x, at.y, W, H, A, R, True))
                    j += 1
                else:
                    Path.Log.warning(
                        "Tag[%d] (%.2f, %.2f, %.2f) is too far away to copy: %.2f (%.2f)"
                        % (i, pos.x, pos.y, self.minZ, dist[0], W)
                    )
            else:
                Path.Log.info("tag[%d]: not enabled, skipping" % i)
        print("copied %d tags" % len(tags))
        return tags

    def processEdge(
        self,
        index,
        edge,
        currentLength,
        lastTagLength,
        tagDistance,
        minLength,
        edgeDict,
    ):
        tagCount = 0
        currentLength += edge.Length
        if edge.Length >= minLength:
            while lastTagLength + tagDistance < currentLength:
                tagCount += 1
                lastTagLength += tagDistance
            if tagCount > 0:
                Path.Log.debug("      index=%d -> count=%d" % (index, tagCount))
                edgeDict[index] = tagCount
        else:
            Path.Log.debug("      skipping=%-2d (%.2f)" % (index, edge.Length))

        return (currentLength, lastTagLength)

    def defaultTagHeight(self):
        op = PathDressup.baseOp(self.obj.Base)
        if hasattr(op, "StartDepth") and hasattr(op, "FinalDepth"):
            pathHeight = (op.StartDepth - op.FinalDepth).Value
        else:
            pathHeight = self.maxZ - self.minZ
        height = HoldingTagPreferences.defaultHeight(pathHeight / 2)
        if height > pathHeight:
            return pathHeight
        return height

    def defaultTagWidth(self):
        width = self.shortestAndLongestPathEdge()[1].Length / 10
        return HoldingTagPreferences.defaultWidth(width)

    def defaultTagAngle(self):
        return HoldingTagPreferences.defaultAngle()

    def defaultTagRadius(self):
        return HoldingTagPreferences.defaultRadius()

    def sortedTags(self, tags):
        ordered = []
        for edge in self.bottomEdges:
            ts = [
                t
                for t in tags
                if Path.Geom.isRoughly(
                    0, Part.Vertex(t.originAt(self.minZ)).distToShape(edge)[0], 0.1
                )
            ]
            for t in sorted(
                ts,
                key=lambda t, edge=edge: (
                    t.originAt(self.minZ) - edge.valueAt(edge.FirstParameter)
                ).Length,
            ):
                tags.remove(t)
                ordered.append(t)
        # disable all tags that are not on the base wire.
        for tag in tags:
            Path.Log.info(
                "Tag #%d (%.2f, %.2f, %.2f) not on base wire - disabling\n"
                % (len(ordered), tag.x, tag.y, self.minZ)
            )
            tag.enabled = False
            ordered.append(tag)
        return ordered

    def pointIsOnPath(self, p):
        v = Part.Vertex(self.pointAtBottom(p))
        Path.Log.debug("pt = (%f, %f, %f)" % (v.X, v.Y, v.Z))
        for e in self.bottomEdges:
            indent = "{} ".format(e.distToShape(v)[0])
            debugEdge(e, indent, True)
            if Path.Geom.isRoughly(0.0, v.distToShape(e)[0], 0.1):
                return True
        return False

    def pointAtBottom(self, p):
        return FreeCAD.Vector(p.x, p.y, self.minZ)


class ObjectTagDressup:
    def __init__(self, obj, base):

        obj.addProperty(
            "App::PropertyLink",
            "Base",
            "Base",
            QT_TRANSLATE_NOOP("App::Property", "The base path to modify"),
        )
        obj.addProperty(
            "App::PropertyLength",
            "Width",
            "Tag",
            QT_TRANSLATE_NOOP("App::Property", "Width of tags."),
        )
        obj.addProperty(
            "App::PropertyLength",
            "Height",
            "Tag",
            QT_TRANSLATE_NOOP("App::Property", "Height of tags."),
        )
        obj.addProperty(
            "App::PropertyAngle",
            "Angle",
            "Tag",
            QT_TRANSLATE_NOOP("App::Property", "Angle of tag plunge and ascent."),
        )
        obj.addProperty(
            "App::PropertyLength",
            "Radius",
            "Tag",
            QT_TRANSLATE_NOOP("App::Property", "Radius of the fillet for the tag."),
        )
        obj.addProperty(
            "App::PropertyVectorList",
            "Positions",
            "Tag",
            QT_TRANSLATE_NOOP("App::Property", "Locations of inserted holding tags"),
        )
        obj.addProperty(
            "App::PropertyIntegerList",
            "Disabled",
            "Tag",
            QT_TRANSLATE_NOOP("App::Property", "IDs of disabled holding tags"),
        )
        obj.addProperty(
            "App::PropertyInteger",
            "SegmentationFactor",
            "Tag",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Factor determining the # of segments used to approximate rounded tags.",
            ),
        )

        self.obj = obj
        self.solids = []
        self.tags = []
        self.pathData = None
        self.toolRadius = None
        self.mappers = []

        obj.Proxy = self
        obj.Base = base

    def dumps(self):
        return None

    def loads(self, state):
        self.obj = state
        self.solids = []
        self.tags = []
        self.pathData = None
        self.toolRadius = None
        self.mappers = []
        return None

    def onChanged(self, obj, prop):
        if prop == "Path" and obj.ViewObject:
            obj.ViewObject.signalChangeIcon()

    def onDocumentRestored(self, obj):
        self.obj = obj

    def supportsTagGeneration(self, obj):
        if not self.pathData:
            self.setup(obj)
        return self.pathData.supportsTagGeneration()

    def generateTags(self, obj, count):
        if self.supportsTagGeneration(obj):
            if self.pathData:
                self.tags = self.pathData.generateTags(
                    obj,
                    count,
                    obj.Width.Value,
                    obj.Height.Value,
                    obj.Angle,
                    obj.Radius.Value,
                    None,
                )
                obj.Positions = [tag.originAt(self.pathData.minZ) for tag in self.tags]
                obj.Disabled = []
                return False
            else:
                self.setup(obj, count)
                self.execute(obj)
                return True
        else:
            self.tags = []
            obj.Positions = []
            obj.Disabled = []
            return False

    def copyTags(self, obj, fromObj):
        obj.Width = fromObj.Width
        obj.Height = fromObj.Height
        obj.Angle = fromObj.Angle
        obj.Radius = fromObj.Radius
        obj.SegmentationFactor = fromObj.SegmentationFactor

        self.tags = self.pathData.copyTags(
            obj, fromObj, obj.Width.Value, obj.Height.Value, obj.Angle, obj.Radius.Value
        )
        obj.Positions = [tag.originAt(self.pathData.minZ) for tag in self.tags]
        obj.Disabled = []
        return False

    def isValidTagStartIntersection(self, edge, i):
        if Path.Geom.pointsCoincide(i, edge.valueAt(edge.LastParameter)):
            return False
        p1 = edge.valueAt(edge.FirstParameter)
        p2 = edge.valueAt(edge.LastParameter)
        if Path.Geom.pointsCoincide(Path.Geom.xy(p1), Path.Geom.xy(p2)):
            # if this vertical goes up, it can't be the start of a tag intersection
            if p1.z < p2.z:
                return False
        return True

    def createPath(self, obj, pathData, tags):
        Path.Log.track()
        commands = []
        lastEdge = 0
        lastTag = 0
        t = 0
        edge = None

        segm = 50
        if hasattr(obj, "SegmentationFactor"):
            segm = obj.SegmentationFactor
            if segm <= 0:
                segm = 50
                obj.SegmentationFactor = 50

        self.mappers = []
        mapper = None

        tc = PathDressup.toolController(obj.Base)
        horizFeed = tc.HorizFeed.Value
        vertFeed = tc.VertFeed.Value
        horizRapid = tc.HorizRapid.Value
        vertRapid = tc.VertRapid.Value

        while edge or lastEdge < len(pathData.edges):
            Path.Log.debug("------- lastEdge = %d/%d.%d/%d" % (lastEdge, lastTag, t, len(tags)))
            if not edge:
                edge = pathData.edges[lastEdge]
                debugEdge(edge, "=======  new edge: %d/%d" % (lastEdge, len(pathData.edges)))
                lastEdge += 1

            if mapper:
                mapper.add(edge)
                if mapper.mappingComplete():
                    commands.extend(mapper.commands)
                    edge = mapper.tail
                    mapper = None
                else:
                    edge = None

            if edge:
                tIndex = (t + lastTag) % len(tags)
                t += 1
                i = tags[tIndex].intersects(edge, edge.FirstParameter)
                if i and self.isValidTagStartIntersection(edge, i):
                    mapper = MapWireToTag(
                        edge,
                        tags[tIndex],
                        i,
                        segm,
                        pathData.maxZ,
                        hSpeed=horizFeed,
                        vSpeed=vertFeed,
                    )
                    self.mappers.append(mapper)
                    edge = mapper.tail

            if not mapper and t >= len(tags):
                # gone through all tags, consume edge and move on
                if edge:
                    debugEdge(edge, "++++++++")
                    if pathData.rapid.isRapid(edge):
                        v = edge.Vertexes[1]
                        if (
                            not commands
                            and Path.Geom.isRoughly(0, v.X)
                            and Path.Geom.isRoughly(0, v.Y)
                            and not Path.Geom.isRoughly(0, v.Z)
                        ):
                            # The very first move is just to move to ClearanceHeight
                            commands.append(Path.Command("G0", {"Z": v.Z, "F": horizRapid}))
                        else:
                            commands.append(
                                Path.Command("G0", {"X": v.X, "Y": v.Y, "Z": v.Z, "F": vertRapid})
                            )
                    else:
                        commands.extend(
                            Path.Geom.cmdsForEdge(
                                edge, segm=segm, hSpeed=horizFeed, vSpeed=vertFeed
                            )
                        )
                edge = None
                t = 0

        return Path.Path(commands)

    def problems(self):
        return list([m for m in self.mappers if m.haveProblem])

    def createTagsPositionDisabled(self, obj, positionsIn, disabledIn):
        rawTags = []
        for i, pos in enumerate(positionsIn):
            tag = Tag(
                i,
                pos.x,
                pos.y,
                obj.Width.Value,
                obj.Height.Value,
                obj.Angle,
                obj.Radius,
                i not in disabledIn,
            )
            tag.createSolidsAt(self.pathData.minZ, self.toolRadius)
            rawTags.append(tag)
        # disable all tags that intersect with their previous tag
        prev = None
        tags = []
        positions = []
        disabled = []
        for i, tag in enumerate(self.pathData.sortedTags(rawTags)):
            if tag.enabled:
                if prev:
                    if prev.solid.common(tag.solid).Faces:
                        Path.Log.info("Tag #%d intersects with previous tag - disabling\n" % i)
                        Path.Log.debug("this tag = %d [%s]" % (i, tag.solid.BoundBox))
                        tag.enabled = False
                elif self.pathData.edges:
                    e = self.pathData.edges[0]
                    p0 = e.valueAt(e.FirstParameter)
                    p1 = e.valueAt(e.LastParameter)
                    if tag.solid.isInside(p0, Path.Geom.Tolerance, True) or tag.solid.isInside(
                        p1, Path.Geom.Tolerance, True
                    ):
                        Path.Log.info("Tag #%d intersects with starting point - disabling\n" % i)
                        tag.enabled = False

            if tag.enabled:
                prev = tag
                Path.Log.debug("previousTag = %d [%s]" % (i, prev))
            else:
                disabled.append(i)
            tag.nr = i  # assign final nr
            tags.append(tag)
            positions.append(tag.originAt(self.pathData.minZ))
        return (tags, positions, disabled)

    def execute(self, obj):
        # import cProfile
        # pr = cProfile.Profile()
        # pr.enable()
        self.doExecute(obj)
        # pr.disable()
        # pr.print_stats()

    def doExecute(self, obj):
        if not obj.Base:
            return
        if not obj.Base.isDerivedFrom("Path::Feature"):
            return
        if not obj.Base.Path:
            return
        if not obj.Base.Path.Commands:
            return

        pathData = self.setup(obj)
        if not pathData:
            Path.Log.debug("execute - no pathData")
            return

        self.tags = []
        if hasattr(obj, "Positions"):
            self.tags, positions, disabled = self.createTagsPositionDisabled(
                obj, obj.Positions, obj.Disabled
            )
            if obj.Disabled != disabled:
                Path.Log.debug("Updating properties.... %s vs. %s" % (obj.Disabled, disabled))
                obj.Positions = positions
                obj.Disabled = disabled

        if not self.tags:
            Path.Log.debug("execute - no tags")
            obj.Path = PathUtils.getPathWithPlacement(obj.Base)
            return

        try:
            self.processTags(obj)
        except Exception as e:
            Path.Log.error("processing tags failed clearing all tags… '%s'" % (e.args[0]))
            obj.Path = PathUtils.getPathWithPlacement(obj.Base)

        # update disabled in case there are some additional ones
        disabled = copy.copy(self.obj.Disabled)
        solids = []
        for tag in self.tags:
            solids.append(tag.solid)
            if not tag.enabled and tag.nr not in disabled:
                disabled.append(tag.nr)
        self.solids = solids
        if obj.Disabled != disabled:
            obj.Disabled = disabled

    @waiting_effects
    def processTags(self, obj):
        tagID = 0
        if Path.Log.getLevel(Path.Log.thisModule()) == Path.Log.Level.DEBUG:
            for tag in self.tags:
                tagID += 1
                if tag.enabled:
                    Path.Log.debug("x=%s, y=%s, z=%s" % (tag.x, tag.y, self.pathData.minZ))
                    # debugMarker(FreeCAD.Vector(tag.x, tag.y, self.pathData.minZ), "tag-%02d" % tagID , (1.0, 0.0, 1.0), 0.5)
                    # if not Path.Geom.isRoughly(90, tag.angle):
                    #    debugCone(tag.originAt(self.pathData.minZ), tag.r1, tag.r2, tag.actualHeight, "tag-%02d" % tagID)
                    # else:
                    #    debugCylinder(tag.originAt(self.pathData.minZ), tag.fullWidth()/2, tag.actualHeight, "tag-%02d" % tagID)

        obj.Path = self.createPath(obj, self.pathData, self.tags)

    def setup(self, obj, generate=False):
        Path.Log.debug("setup")
        self.obj = obj
        try:
            pathData = PathData(obj)
        except ValueError:
            Path.Log.error(
                translate(
                    "CAM_DressupTag",
                    "Cannot insert holding tags for this path - select a profile path",
                )
                + "\n"
            )
            return None

        self.toolRadius = float(PathDressup.toolController(obj.Base).Tool.Diameter) / 2
        self.pathData = pathData
        if generate:
            obj.Height = self.pathData.defaultTagHeight()
            obj.Width = self.pathData.defaultTagWidth()
            obj.Angle = self.pathData.defaultTagAngle()
            obj.Radius = self.pathData.defaultTagRadius()
            count = HoldingTagPreferences.defaultCount()
            self.generateTags(obj, count)
        return self.pathData

    def setXyEnabled(self, triples):
        Path.Log.track()
        if not self.pathData:
            self.setup(self.obj)
        positions = []
        disabled = []
        for i, (x, y, enabled) in enumerate(triples):
            # print("%d: (%.2f, %.2f) %d" % (i, x, y, enabled))
            positions.append(FreeCAD.Vector(x, y, 0))
            if not enabled:
                disabled.append(i)
        (
            self.tags,
            self.obj.Positions,
            self.obj.Disabled,
        ) = self.createTagsPositionDisabled(self.obj, positions, disabled)
        self.processTags(self.obj)

    def pointIsOnPath(self, obj, point):
        if not self.pathData:
            self.setup(obj)
        return self.pathData.pointIsOnPath(point)

    def pointAtBottom(self, obj, point):
        if not self.pathData:
            self.setup(obj)
        return self.pathData.pointAtBottom(point)


def Create(baseObject, name="DressupTag"):
    """
    Create(basePath, name='DressupTag') … create tag dressup object for the given base path.
    """
    if not baseObject.isDerivedFrom("Path::Feature"):
        Path.Log.error(translate("CAM_DressupTag", "The selected object is not a path") + "\n")
        return None

    if baseObject.isDerivedFrom("Path::FeatureCompoundPython"):
        Path.Log.error(translate("CAM_DressupTag", "Select a profile object"))
        return None

    obj = FreeCAD.ActiveDocument.addObject("Path::FeaturePython", name)
    dbo = ObjectTagDressup(obj, baseObject)
    job = PathUtils.findParentJob(baseObject)
    job.Proxy.addOperation(obj, baseObject)
    dbo.setup(obj, True)
    return obj


Path.Log.notice("Loading CAM_DressupTag… done\n")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 FreeCAD Developers                                 *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides functions to create offsets of different shapes."""
## @package offset
# \ingroup draftfunctions
# \brief Provides functions to create offsets of different shapes.

## \addtogroup draftfunctions
# @{
import math

import FreeCAD as App
import DraftVecUtils
from draftutils import gui_utils
from draftutils import params
from draftutils import utils

from draftmake.make_rectangle import make_rectangle
from draftmake.make_wire import make_wire
from draftmake.make_polygon import make_polygon
from draftmake.make_circle import make_circle
from draftmake.make_bspline import make_bspline


def offset(obj, delta, copy=False, bind=False, sym=False, occ=False):
    """offset(object,delta,[copymode],[bind])

    Offset the given wire by applying the given delta Vector to its first
    vertex.

    Parameters
    ----------
    obj :

    delta : Base.Vector or list of Base.Vector
        If offsetting a BSpline, the delta must not be a Vector but a list
        of Vectors, one for each node of the spline.

    copy : bool
        If copymode is True, another object is created, otherwise the same
        object gets offsetted.

    copy : bool
        If bind is True, and provided the wire is open, the original
        and the offset wires will be bound by their endpoints, forming a face.

    sym : bool
        if sym is True, bind must be true too, and the offset is made on both
        sides, the total width being the given delta length.
    """
    import Part
    import DraftGeomUtils

    newwire = None
    delete = None

    if copy is False and (
        utils.get_type(obj).startswith("Sketcher::")
        or utils.get_type(obj).startswith("Part::")
        or utils.get_type(obj).startswith("PartDesign::")
    ):  # For PartDesign_SubShapeBinders which can reference sketches.
        print(
            "the offset tool is currently unable to offset a non-Draft object directly - Creating a copy"
        )
        copy = True

    def getRect(p, obj):
        """returns length,height,placement"""
        pl = obj.Placement.copy()
        pl.Base = p[0]
        diag = p[2].sub(p[0])
        bb = p[1].sub(p[0])
        bh = p[3].sub(p[0])
        nb = DraftVecUtils.project(diag, bb)
        nh = DraftVecUtils.project(diag, bh)
        if obj.Length.Value < 0:
            l = -nb.Length
        else:
            l = nb.Length
        if obj.Height.Value < 0:
            h = -nh.Length
        else:
            h = nh.Length
        return l, h, pl

    def getRadius(obj, delta):
        """returns a new radius for a regular polygon"""
        an = math.pi / obj.FacesNumber
        nr = DraftVecUtils.rotate(delta, -an)
        nr.multiply(1 / math.cos(an))
        nr = obj.Shape.Vertexes[0].Point.add(nr)
        nr = nr.sub(obj.Placement.Base)
        nr = nr.Length
        if obj.DrawMode == "inscribed":
            return nr
        else:
            return nr * math.cos(math.pi / obj.FacesNumber)

    newwire = None
    if utils.get_type(obj) == "Circle":
        pass
    elif utils.get_type(obj) == "BSpline":
        pass
    else:
        if sym:
            d1 = App.Vector(delta).multiply(0.5)
            d2 = d1.negative()
            n1 = DraftGeomUtils.offsetWire(obj.Shape, d1)
            n2 = DraftGeomUtils.offsetWire(obj.Shape, d2)
        else:
            if isinstance(delta, float) and (len(obj.Shape.Edges) == 1):
                # circle
                c = obj.Shape.Edges[0].Curve
                nc = Part.Circle(c.Center, c.Axis, delta)
                if len(obj.Shape.Vertexes) > 1:
                    nc = Part.ArcOfCircle(
                        nc, obj.Shape.Edges[0].FirstParameter, obj.Shape.Edges[0].LastParameter
                    )
                newwire = Part.Wire(nc.toShape())
                p = []
            else:
                newwire = DraftGeomUtils.offsetWire(obj.Shape, delta)
                if DraftGeomUtils.hasCurves(newwire) and copy:
                    p = []
                else:
                    p = DraftGeomUtils.getVerts(newwire)
    if occ:
        newobj = App.ActiveDocument.addObject("Part::Feature", "Offset")
        newobj.Shape = DraftGeomUtils.offsetWire(obj.Shape, delta, occ=True)
        gui_utils.formatObject(newobj, obj)
        if not copy:
            delete = obj.Name
    elif bind:
        if not DraftGeomUtils.isReallyClosed(obj.Shape):
            if sym:
                s1 = n1
                s2 = n2
            else:
                s1 = obj.Shape
                s2 = newwire
            if s1 and s2:
                w1 = s1.Edges
                w2 = s2.Edges
                w3 = Part.LineSegment(s1.Vertexes[0].Point, s2.Vertexes[0].Point).toShape()
                w4 = Part.LineSegment(s1.Vertexes[-1].Point, s2.Vertexes[-1].Point).toShape()
                newobj = App.ActiveDocument.addObject("Part::Feature", "Offset")
                newobj.Shape = Part.Face(Part.Wire(w1 + [w3] + w2 + [w4]))
            else:
                print("Draft.offset: Unable to bind wires")
        else:
            newobj = App.ActiveDocument.addObject("Part::Feature", "Offset")
            newobj.Shape = Part.Face(obj.Shape.Wires[0])
        if not copy:
            delete = obj.Name
    elif copy:
        newobj = None
        if sym:
            return None
        if utils.get_type(obj) == "Wire":
            if p:
                newobj = make_wire(p)
                newobj.Closed = obj.Closed
            elif newwire:
                newobj = App.ActiveDocument.addObject("Part::Feature", "Offset")
                newobj.Shape = newwire
            else:
                print("Draft.offset: Unable to duplicate this object")
        elif utils.get_type(obj) == "Rectangle":
            if p:
                length, height, plac = getRect(p, obj)
                newobj = make_rectangle(length, height, plac)
            elif newwire:
                newobj = App.ActiveDocument.addObject("Part::Feature", "Offset")
                newobj.Shape = newwire
            else:
                print("Draft.offset: Unable to duplicate this object")
        elif utils.get_type(obj) == "Circle":
            pl = obj.Placement
            newobj = make_circle(delta)
            newobj.FirstAngle = obj.FirstAngle
            newobj.LastAngle = obj.LastAngle
            newobj.Placement = pl
        elif utils.get_type(obj) == "Polygon":
            pl = obj.Placement
            newobj = make_polygon(obj.FacesNumber)
            newobj.Radius = getRadius(obj, delta)
            newobj.DrawMode = obj.DrawMode
            newobj.Placement = pl
        elif utils.get_type(obj) == "BSpline":
            newobj = make_bspline(delta)
            newobj.Closed = obj.Closed
        else:
            # try to offset anyway
            try:
                if p:
                    newobj = make_wire(p)
                    newobj.Closed = DraftGeomUtils.isReallyClosed(obj.Shape)
            except Part.OCCError:
                pass
            if (not newobj) and newwire:
                newobj = App.ActiveDocument.addObject("Part::Feature", "Offset")
                newobj.Shape = newwire
            if not newobj:
                print("Draft.offset: Unable to create an offset")
        if newobj:
            gui_utils.formatObject(newobj, obj)
    else:
        newobj = None
        if sym:
            return None
        if utils.get_type(obj) == "Wire":
            if obj.Base or obj.Tool:
                App.Console.PrintWarning("Warning: object history removed\n")
                obj.Base = None
                obj.Tool = None
            obj.Placement = App.Placement()  # p points are in the global coordinate system
            obj.Points = p
        elif utils.get_type(obj) == "BSpline":
            # print(delta)
            obj.Points = delta
            # print("done")
        elif utils.get_type(obj) == "Rectangle":
            length, height, plac = getRect(p, obj)
            obj.Placement = plac
            obj.Length = length
            obj.Height = height
        elif utils.get_type(obj) == "Circle":
            obj.Radius = delta
        elif utils.get_type(obj) == "Polygon":
            obj.Radius = getRadius(obj, delta)
        elif utils.get_type(obj) == "Part":
            print("unsupported object")  # TODO
        newobj = obj
    if copy and params.get_param("selectBaseObjects"):
        gui_utils.select(obj)
    else:
        gui_utils.select(newobj)
    if delete:
        App.ActiveDocument.removeObject(delete)
    return newobj


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2024 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""BIM Roof command"""

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate

PARAMS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/BIM")


class Arch_Roof:
    """the Arch Roof command definition"""

    def GetResources(self):
        return {
            "Pixmap": "Arch_Roof",
            "MenuText": QT_TRANSLATE_NOOP("Arch_Roof", "Roof"),
            "Accel": "R, F",
            "ToolTip": QT_TRANSLATE_NOOP(
                "Arch_Roof", "Creates a roof object from the selected wire."
            ),
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):
        import ArchComponent

        sel = FreeCADGui.Selection.getSelectionEx()
        if sel:
            sel = sel[0]
            obj = sel.Object
            FreeCADGui.Control.closeDialog()
            if sel.HasSubObjects:
                if "Face" in sel.SubElementNames[0]:
                    i = int(sel.SubElementNames[0][4:])
                    FreeCAD.ActiveDocument.openTransaction(translate("Arch", "Create Roof"))
                    FreeCADGui.addModule("Arch")
                    FreeCADGui.doCommand(
                        "obj = Arch.makeRoof(FreeCAD.ActiveDocument."
                        + obj.Name
                        + ","
                        + str(i)
                        + ")"
                    )
                    FreeCADGui.addModule("Draft")
                    FreeCADGui.doCommand("Draft.autogroup(obj)")
                    FreeCAD.ActiveDocument.commitTransaction()
                    FreeCAD.ActiveDocument.recompute()
                    return
            if hasattr(obj, "Shape"):
                if obj.Shape.Wires:
                    FreeCAD.ActiveDocument.openTransaction(translate("Arch", "Create Roof"))
                    FreeCADGui.addModule("Arch")
                    FreeCADGui.doCommand(
                        "obj = Arch.makeRoof(FreeCAD.ActiveDocument." + obj.Name + ")"
                    )
                    FreeCADGui.addModule("Draft")
                    FreeCADGui.doCommand("Draft.autogroup(obj)")
                    FreeCAD.ActiveDocument.commitTransaction()
                    FreeCAD.ActiveDocument.recompute()
                    return
            else:
                FreeCAD.Console.PrintMessage(translate("Arch", "Unable to create a roof"))
        else:
            FreeCAD.Console.PrintMessage(translate("Arch", "Please select a base object") + "\n")
            FreeCADGui.Control.showDialog(ArchComponent.SelectionTaskPanel())
            FreeCAD.ArchObserver = ArchComponent.ArchSelectionObserver(nextCommand="Arch_Roof")
            FreeCADGui.Selection.addObserver(FreeCAD.ArchObserver)


FreeCADGui.addCommand("Arch_Roof", Arch_Roof())



# SPDX-License-Identifier: LGPL-2.1-or-later

# /***************************************************************************
# *   Copyright (c) 2024 Eric Price (CorvusCorax)                           *
# *                      <eric.price[at]tuebingen.mpg.de>                   *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# ***************************************************************************/

__title__ = "BOPTools.ToleranceFeatures module"
__author__ = "CorvusCorax"
__url__ = "https://www.freecad.org"
__doc__ = "Implementation of document objects (features) to adjust/manipulate tolerances."

import FreeCAD
import Part

if FreeCAD.GuiUp:
    import FreeCADGui
    from PySide import QtCore, QtGui

    # -------------------------- common stuff -------------------------------------

    # -------------------------- translation-related code -------------------------

    try:
        _fromUtf8 = QtCore.QString.fromUtf8
    except Exception:

        def _fromUtf8(s):
            return s

    translate = FreeCAD.Qt.translate
# --------------------------/translation-related code -------------------------


def getParamRefine():
    return FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Part/Boolean").GetBool(
        "RefineModel"
    )


def cmdCreateToleranceSetFeature(name, minTolerance=1e-7, maxTolerance=0):
    """cmdCreateToleranceSetFeature(name, minTolerance, maxTolerance): generalized implementation of GUI commands."""
    sel = FreeCADGui.Selection.getSelectionEx()

    FreeCAD.ActiveDocument.openTransaction("Create ToleranceSet")
    FreeCADGui.addModule("BOPTools.ToleranceFeatures")
    FreeCADGui.doCommand(
        "j = BOPTools.ToleranceFeatures.makeToleranceSet(name='{name}')".format(name=name)
    )
    FreeCADGui.doCommand("j.minTolerance = {minTolerance}".format(minTolerance=minTolerance))
    FreeCADGui.doCommand("j.maxTolerance = {maxTolerance}".format(maxTolerance=maxTolerance))
    FreeCADGui.doCommand(
        "j.Objects = {sel}".format(
            sel="[" + ", ".join(["App.ActiveDocument." + so.Object.Name for so in sel]) + "]"
        )
    )

    try:
        FreeCADGui.doCommand("j.Proxy.execute(j)")
        FreeCADGui.doCommand("j.purgeTouched()")
    except Exception as err:
        mb = QtGui.QMessageBox()
        mb.setIcon(mb.Icon.Warning)
        error_text1 = translate(
            "Part_ToleranceFeatures", "Computing the result failed with an error:"
        )
        error_text2 = translate(
            "Part_ToleranceFeatures",
            "Click 'Continue' to create the feature anyway, or 'Abort' to cancel.",
        )
        mb.setText(error_text1 + "\n\n" + str(err) + "\n\n" + error_text2)
        mb.setWindowTitle(translate("Part_ToleranceFeatures", "Bad Selection", None))
        btnAbort = mb.addButton(QtGui.QMessageBox.StandardButton.Abort)
        btnOK = mb.addButton(
            translate("Part_ToleranceFeatures", "Continue", None),
            QtGui.QMessageBox.ButtonRole.ActionRole,
        )
        mb.setDefaultButton(btnOK)
        mb.exec_()

        if mb.clickedButton() is btnAbort:
            FreeCAD.ActiveDocument.abortTransaction()
            return

    FreeCADGui.doCommand(
        "for obj in j.ViewObject.Proxy.claimChildren():\n" "    obj.ViewObject.hide()"
    )

    FreeCAD.ActiveDocument.commitTransaction()


def getIconPath(icon_dot_svg):
    return icon_dot_svg


# -------------------------- /common stuff ------------------------------------

# -------------------------- Connect ------------------------------------------


def makeToleranceSet(name):
    """makeToleranceSet(name): makes an ToleranceSet object."""
    obj = FreeCAD.ActiveDocument.addObject("Part::FeaturePython", name)
    FeatureToleranceSet(obj)
    if FreeCAD.GuiUp:
        ViewProviderToleranceSet(obj.ViewObject)
    return obj


class FeatureToleranceSet:
    """The PartToleranceSetFeature object."""

    def __init__(self, obj):
        obj.addProperty(
            "App::PropertyLinkList",
            "Objects",
            "ToleranceSet",
            "Objects to have tolerance adjusted.",
            locked=True,
        )
        obj.addProperty(
            "App::PropertyBool",
            "Refine",
            "ToleranceSet",
            "True = refine resulting shape. False = output as is.",
            locked=True,
        )
        obj.addProperty(
            "App::PropertyLength", "minTolerance", "ToleranceSet", "0.1 nm", locked=True
        )
        obj.addProperty("App::PropertyLength", "maxTolerance", "ToleranceSet", "0", locked=True)
        obj.Refine = getParamRefine()

        obj.Proxy = self
        self.Type = "FeatureToleranceSet"

    def onDocumentRestored(self, obj):
        if not hasattr(obj, "maxTolerance"):
            obj.addProperty("App::PropertyLength", "maxTolerance", "ToleranceSet", "0", locked=True)

    def execute(self, selfobj):
        shapes = []
        for obj in selfobj.Objects:
            sh = obj.Shape.copy(True, False)
            sh.limitTolerance(selfobj.minTolerance, selfobj.maxTolerance)
            if selfobj.Refine:
                sh.fix(selfobj.minTolerance, selfobj.minTolerance, selfobj.maxTolerance)
                sh = sh.removeSplitter()
                sh.fix(selfobj.minTolerance, selfobj.minTolerance, selfobj.maxTolerance)
            shapes.append(sh)

        if len(shapes) > 1:
            rst = Part.makeCompound(shapes)
        else:
            rst = shapes[0]
        selfobj.Shape = rst


class ViewProviderToleranceSet:
    """A View Provider for the Part ToleranceSet feature."""

    def __init__(self, vobj):
        vobj.Proxy = self

    def getIcon(self):
        return ":/icons/preferences-part_design.svg"

    def attach(self, vobj):
        self.ViewObject = vobj
        self.Object = vobj.Object

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def claimChildren(self):
        return self.Object.Objects

    def onDelete(self, feature, subelements):
        try:
            for obj in self.claimChildren():
                obj.ViewObject.show()
        except Exception as err:
            FreeCAD.Console.PrintError("Error in onDelete: " + str(err))
        return True

    def canDragObjects(self):
        return True

    def canDropObjects(self):
        return True

    def canDragObject(self, dragged_object):
        return True

    def canDropObject(self, incoming_object):
        return hasattr(incoming_object, "Shape")

    def dragObject(self, selfvp, dragged_object):
        objs = self.Object.Objects
        objs.remove(dragged_object)
        self.Object.Objects = objs

    def dropObject(self, selfvp, incoming_object):
        self.Object.Objects = self.Object.Objects + [incoming_object]


class CommandToleranceSet:
    """Command to create ToleranceSet feature."""

    def GetResources(self):
        return {
            "Pixmap": getIconPath("preferences-part_design.svg"),
            "MenuText": QtCore.QT_TRANSLATE_NOOP("Part_ToleranceSet", "Set Tolerance"),
            "Accel": "",
            "ToolTip": QtCore.QT_TRANSLATE_NOOP(
                "Part_ToleranceSet",
                "Creates a parametric copy of the selected object with all contained tolerances set to at least a certain minimum value",
            ),
        }

    def Activated(self):
        if len(FreeCADGui.Selection.getSelectionEx()) >= 1:
            cmdCreateToleranceSetFeature(name="Tolerance")
        else:
            mb = QtGui.QMessageBox()
            mb.setIcon(mb.Icon.Warning)
            mb.setText(
                translate("Part_ToleranceSet", "Select at least one object or compounds", None)
            )
            mb.setWindowTitle(translate("Part_ToleranceSet", "Bad Selection", None))
            mb.exec_()

    def IsActive(self):
        if FreeCAD.ActiveDocument:
            return True
        else:
            return False


# -------------------------- /Connect -----------------------------------------


def addCommands():
    FreeCADGui.addCommand("Part_ToleranceSet", CommandToleranceSet())



# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""A LazyLoader class."""

from __future__ import absolute_import
from __future__ import division

import importlib
import types


class LazyLoader(types.ModuleType):
  """Lazily import a module, mainly to avoid pulling in large dependencies.

  `contrib`, and `ffmpeg` are examples of modules that are large and not always
  needed, and this allows them to only be loaded when they are used.
  """

  # The lint error here is incorrect.
  def __init__(self, local_name, parent_module_globals, name, warning=None):  # pylint: disable=super-on-old-class
    self._local_name = local_name
    self._parent_module_globals = parent_module_globals
    self._warning = warning

    super(LazyLoader, self).__init__(name)

  def _load(self):
    """Load the module and insert it into the parent's globals."""
    # Import the target module and insert it into the parent's namespace
    module = importlib.import_module(self.__name__)
    self._parent_module_globals[self._local_name] = module

    # Update this object's dict so that if someone keeps a reference to the
    #   LazyLoader, lookups are efficient (__getattr__ is only called on lookups
    #   that fail).
    self.__dict__.update(module.__dict__)

    return module

  def __getattr__(self, item):
    module = self._load()
    return getattr(module, item)

  def __dir__(self):
    module = self._load()
    return dir(module)


# SPDX-License-Identifier: LGPL-2.1-or-later
# /**************************************************************************
#                                                                           *
#    Copyright (c) 2023 Ondsel <development@ondsel.com>                     *
#                                                                           *
#    This file is part of FreeCAD.                                          *
#                                                                           *
#    FreeCAD is free software: you can redistribute it and/or modify it     *
#    under the terms of the GNU Lesser General Public License as            *
#    published by the Free Software Foundation, either version 2.1 of the   *
#    License, or (at your option) any later version.                        *
#                                                                           *
#    FreeCAD is distributed in the hope that it will be useful, but         *
#    WITHOUT ANY WARRANTY; without even the implied warranty of             *
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
#    Lesser General Public License for more details.                        *
#                                                                           *
#    You should have received a copy of the GNU Lesser General Public       *
#    License along with FreeCAD. If not, see                                *
#    <https://www.gnu.org/licenses/>.                                       *
#                                                                           *
# **************************************************************************/

import Assembly_rc


class AssemblyCommandGroup:
    def __init__(self, cmdlist, menu, tooltip=None):
        self.cmdlist = cmdlist
        self.menu = menu
        if tooltip is None:
            self.tooltip = menu
        else:
            self.tooltip = tooltip

    def GetCommands(self):
        return tuple(self.cmdlist)

    def GetResources(self):
        return {"MenuText": self.menu, "ToolTip": self.tooltip}

    def IsActive(self):
        if FreeCAD.ActiveDocument is not None:
            return True
        return False


class AssemblyWorkbench(Workbench):
    "Assembly workbench"

    def __init__(self):
        self.__class__.Icon = (
            FreeCAD.getResourceDir() + "Mod/Assembly/Resources/icons/AssemblyWorkbench.svg"
        )
        self.__class__.MenuText = "Assembly"
        self.__class__.ToolTip = "Assembly workbench"

    def Initialize(self):
        global AssemblyCommandGroup

        translate = FreeCAD.Qt.translate

        # load the builtin modules
        from PySide import QtCore, QtGui
        from PySide.QtCore import QT_TRANSLATE_NOOP
        import CommandCreateAssembly, CommandInsertLink, CommandInsertNewPart, CommandCreateJoint, CommandSolveAssembly, CommandExportASMT, CommandCreateView, CommandCreateSimulation, CommandCreateBom
        import Preferences

        FreeCADGui.addLanguagePath(":/translations")
        FreeCADGui.addIconPath(":/icons")

        FreeCADGui.addPreferencePage(
            Preferences.PreferencesPage, QT_TRANSLATE_NOOP("QObject", "Assembly")
        )

        # build commands list
        cmdList = [
            "Assembly_CreateAssembly",
            "Assembly_Insert",
            "Assembly_SolveAssembly",
            "Assembly_CreateView",
            "Assembly_CreateSimulation",
            "Assembly_CreateBom",
        ]

        cmdListMenuOnly = [
            "Assembly_ExportASMT",
        ]

        cmdListJoints = [
            "Assembly_ToggleGrounded",
            "Separator",
            "Assembly_CreateJointFixed",
            "Assembly_CreateJointRevolute",
            "Assembly_CreateJointCylindrical",
            "Assembly_CreateJointSlider",
            "Assembly_CreateJointBall",
            "Separator",
            "Assembly_CreateJointDistance",
            "Assembly_CreateJointParallel",
            "Assembly_CreateJointPerpendicular",
            "Assembly_CreateJointAngle",
            "Separator",
            "Assembly_CreateJointRackPinion",
            "Assembly_CreateJointScrew",
            "Assembly_CreateJointGearBelt",
        ]

        self.appendToolbar(QT_TRANSLATE_NOOP("Workbench", "Assembly"), cmdList)
        self.appendToolbar(QT_TRANSLATE_NOOP("Workbench", "Assembly Joints"), cmdListJoints)

        self.appendMenu(
            [QT_TRANSLATE_NOOP("Workbench", "&Assembly")],
            cmdList + cmdListMenuOnly + ["Separator"] + cmdListJoints,
        )

    def Activated(self):
        # update the translation engine
        FreeCADGui.updateLocale()

        # Add task watchers to provide contextual tools in the task panel
        self.setWatchers()

    def Deactivated(self):
        FreeCADGui.Control.clearTaskWatcher()

    def ContextMenu(self, recipient):
        pass

    def setWatchers(self):
        import UtilsAssembly

        translate = FreeCAD.Qt.translate

        class AssemblyCreateWatcher:
            """Shows 'Create Assembly' when no assembly exists in the document."""

            def __init__(self):
                self.commands = ["Assembly_CreateAssembly"]
                self.title = translate("Assembly", "Create")

            def shouldShow(self):
                doc = FreeCAD.ActiveDocument

                if hasattr(doc, "RootObjects"):
                    for obj in doc.RootObjects:
                        if obj.isDerivedFrom("Assembly::AssemblyObject"):
                            return False
                return True

        class AssemblyActivateWatcher:
            """Shows 'Activate Assembly' when an assembly exists but is not active."""

            def __init__(self):
                self.commands = ["Assembly_ActivateAssembly"]
                self.title = translate("Assembly", "Activate")

            def shouldShow(self):
                doc = FreeCAD.ActiveDocument

                has_assembly = False
                if hasattr(doc, "RootObjects"):
                    for obj in doc.RootObjects:
                        if obj.isDerivedFrom("Assembly::AssemblyObject"):
                            has_assembly = True
                            break

                assembly = UtilsAssembly.activeAssembly()

                return has_assembly and (assembly is None or assembly.Document != doc)

        class AssemblyBaseWatcher:
            """Base class for watchers that require an active assembly."""

            def __init__(self):
                self.assembly = None

            def shouldShow(self):
                doc = FreeCAD.ActiveDocument

                self.assembly = UtilsAssembly.activeAssembly()
                return self.assembly is not None and self.assembly.Document == doc

        class AssemblyInsertWatcher(AssemblyBaseWatcher):
            """Shows 'Insert Component' when an assembly is active."""

            def __init__(self):
                super().__init__()
                self.commands = ["Assembly_Insert"]
                self.title = translate("Assembly", "Insert")

            def shouldShow(self):
                return super().shouldShow()

        class AssemblyGroundWatcher(AssemblyBaseWatcher):
            """Shows 'Ground' when the active assembly has no grounded parts."""

            def __init__(self):
                super().__init__()
                self.commands = ["Assembly_ToggleGrounded"]
                self.title = translate("Assembly", "Grounding")

            def shouldShow(self):
                if not super().shouldShow():
                    return False
                return (
                    UtilsAssembly.assembly_has_at_least_n_parts(1)
                    and not UtilsAssembly.isAssemblyGrounded()
                )

        class AssemblyJointsWatcher(AssemblyBaseWatcher):
            """Shows Joint, View, and BOM tools when there are enough parts."""

            def __init__(self):
                super().__init__()
                self.commands = [
                    "Assembly_CreateJointFixed",
                    "Assembly_CreateJointRevolute",
                    "Assembly_CreateJointCylindrical",
                    "Assembly_CreateJointSlider",
                    "Assembly_CreateJointBall",
                    "Separator",
                    "Assembly_CreateJointDistance",
                    "Assembly_CreateJointParallel",
                    "Assembly_CreateJointPerpendicular",
                    "Assembly_CreateJointAngle",
                ]
                self.title = translate("Assembly", "Constraints")

            def shouldShow(self):
                if not super().shouldShow():
                    return False
                return UtilsAssembly.assembly_has_at_least_n_parts(2)

        class AssemblyToolsWatcher(AssemblyBaseWatcher):
            """Shows Joint, View, and BOM tools when there are enough parts."""

            def __init__(self):
                super().__init__()
                self.commands = [
                    "Assembly_CreateView",
                    "Assembly_CreateBom",
                ]
                self.title = translate("Assembly", "Tools")

            def shouldShow(self):
                if not super().shouldShow():
                    return False
                return UtilsAssembly.assembly_has_at_least_n_parts(1)

        class AssemblySimulationWatcher(AssemblyBaseWatcher):
            """Shows 'Create Simulation' when specific motional joints exist."""

            def __init__(self):
                super().__init__()
                self.commands = ["Assembly_CreateSimulation"]
                self.title = translate("Assembly", "Simulation")

            def shouldShow(self):
                if not super().shouldShow():
                    return False

                joint_types = ["Revolute", "Slider", "Cylindrical"]
                joints = UtilsAssembly.getJointsOfType(self.assembly, joint_types)
                return len(joints) > 0

        watchers = [
            AssemblyCreateWatcher(),
            AssemblyActivateWatcher(),
            AssemblyInsertWatcher(),
            AssemblyGroundWatcher(),
            AssemblyJointsWatcher(),
            AssemblyToolsWatcher(),
            AssemblySimulationWatcher(),
        ]
        FreeCADGui.Control.addTaskWatcher(watchers)


Gui.addWorkbench(AssemblyWorkbench())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Mario Passaglia <mpassaglia[at]cbc.uba.ar>         *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "Tools for the work with Elmer solver"
__author__ = "Mario Passaglia"
__url__ = "https://www.freecad.org"


from PySide.QtCore import QProcess, QProcessEnvironment
import os
import re
import shutil

import FreeCAD

from . import writer
from .. import settings

from femtools import membertools
from femtools.objecttools import ObjectTools


class ElmerTools(ObjectTools):

    name = "Elmer"

    def __init__(self, obj):
        super().__init__(obj)
        self.model_file = ""
        self._result_format = ""

    def prepare(self):
        w = writer.Writer(self.obj, self.obj.WorkingDirectory)
        w.write_solver_input()

        mesh = w.getSingleMember("Fem::FemMeshObject")
        if not mesh.FemMesh.Groups:
            raise ValueError(f"Mesh object '{mesh.Label}' has no groups, please remesh\n")

        mesh_file = os.path.join(self.obj.WorkingDirectory, "mesh.unv")
        mesh.FemMesh.write(mesh_file)

        grid_bin = settings.get_binary("ElmerGrid")
        env = QProcessEnvironment.systemEnvironment()
        p = QProcess()
        p.setProcessEnvironment(env)
        p.setWorkingDirectory(self.obj.WorkingDirectory)
        grid_args = ["8", "2", mesh_file, "-out", self.obj.WorkingDirectory]
        p.start(grid_bin, grid_args)
        p.waitForFinished(-1)
        num_proc = self.fem_param.GetGroup("Elmer").GetInt("NumberOfTasks", 1)
        if num_proc > 1:
            # MPI parallel computing version
            grid_args.extend(["-partdual", "-metiskway", str(num_proc)])
            p.start(grid_bin, grid_args)
            p.waitForFinished()

        self.model_file = os.path.join(self.obj.WorkingDirectory, writer._SIF_NAME)
        handled = w.getHandledConstraints()
        allConstraints = membertools.get_member(self.analysis, "Fem::Constraint")
        for obj in set(allConstraints) - handled:
            FreeCAD.Console.PrintWarning(f"Ignored constraint {obj.Label}")

    def compute(self):
        self._clear_results()
        elmer_bin = settings.get_binary("ElmerSolver")
        num_proc = self.fem_param.GetGroup("Elmer").GetInt("NumberOfTasks", 1)
        num_thr = self.fem_param.GetGroup("Elmer").GetInt("ThreadsPerTask", 1)
        env = QProcessEnvironment.systemEnvironment()
        env.insert("OMP_NUM_THREADS", str(num_thr))
        self.process.setProcessEnvironment(env)
        self.process.setWorkingDirectory(self.obj.WorkingDirectory)

        if num_proc > 1:
            # MPI parallel computing version
            mpi = shutil.which("mpiexec")
            self._result_format = ".pvtu"
            command_list = ["-n", str(num_proc), elmer_bin]
            self.process.start(mpi, command_list)
        else:
            self._result_format = ".vtu"
            command_list = []
            self.process.start(elmer_bin, command_list)

        if self.obj.SimulationType == "Transient":
            self._result_format = ".pvd"

        return self.process

    def update_properties(self):
        self._load_vtk_results()
        self._load_dat_results()

    def _clear_results(self):
        dir_content = os.listdir(self.obj.WorkingDirectory)
        for f in dir_content:
            path = os.path.join(self.obj.WorkingDirectory, f)
            base, ext = os.path.splitext(path)
            if ext in [".vtu", ".vtp", ".pvtu", ".pvd", ".dat"]:
                os.remove(path)

    def _load_vtk_results(self):
        # search current pipeline
        keep_result = self.fem_param.GetGroup("General").GetBool("KeepResultsOnReRun", False)
        pipeline = None
        create = False
        for res in self.obj.Results:
            if res.isDerivedFrom("Fem::FemPostPipeline"):
                pipeline = res

        if not pipeline or keep_result:
            # create pipeline
            pipeline = self.obj.Document.addObject("Fem::FemPostPipeline", self.obj.Name + "Result")
            self.analysis.addObject(pipeline)
            tmp = self.obj.Results
            tmp.append(pipeline)
            self.obj.Results = tmp
            create = True

        files = os.listdir(self.obj.WorkingDirectory)
        for f in files:
            base, ext = os.path.splitext(f)
            if ext == self._result_format:
                res = os.path.join(self.obj.WorkingDirectory, f)
                pipeline.read(res)
                break

        if create:
            # default display mode
            pipeline.ViewObject.DisplayMode = "Surface"
            pipeline.ViewObject.SelectionStyle = "BoundBox"
            fields = pipeline.ViewObject.getEnumerationsOfProperty("Field")
            for f in fields:
                # beware of possible suffix Im or Re
                if f.lower().startswith(self._get_default_field()):
                    pipeline.ViewObject.Field = f
                    break

    def _load_dat_results(self):
        # search dat output
        keep_result = self.fem_param.GetGroup("General").GetBool("KeepResultsOnReRun", False)
        dat = None
        for res in self.obj.Results:
            if res.isDerivedFrom("App::TextDocument"):
                dat = res

        if not dat or keep_result:
            # create dat output
            dat = self.obj.Document.addObject("App::TextDocument", self.obj.Name + "Output")
            self.analysis.addObject(dat)
            tmp = self.obj.Results
            tmp.append(dat)
            self.obj.Results = tmp

        files = os.listdir(self.obj.WorkingDirectory)
        for f in files:
            if f.endswith(".dat"):
                dat_file = os.path.join(self.obj.WorkingDirectory, f)
                with open(dat_file, "r") as file:
                    dat.Text = file.read()
                break

    def _get_default_field(self):
        default = "None"
        for eq in self.obj.Group:
            match eq.Proxy.Type:
                case "Fem::EquationElmerHeat":
                    default = "temperature"
                case "Fem::EquationElmerElasticity" | "Fem::EquationElmerDeformation":
                    default = "displacement"
                case "Fem::EquationElmerElectrostatic" | "Fem::EquationElmerStaticCurrent":
                    default = "potential"
                case "Fem::EquationElmerFlow":
                    default = "pressure"
                case "Fem::EquationElmerMagnetodynamic" | "Fem::EquationElmerMagnetodynamic2D":
                    default = "magnetic flux"
        return default

    def version(self):
        p = QProcess()
        elmer_bin = settings.get_binary("ElmerSolver")
        p.start(elmer_bin, ["-v"])
        p.waitForFinished()
        info = p.readAll().data().decode()
        reg_exp = re.compile(r"Version:\s*(?P<version>.*)$", re.M)
        m = reg_exp.search(info)
        ver = "Version: {}".format(m.group("version") if m else "")
        return ver
